'''
Copyright (C) 2010-2021 Alibaba Group Holding Limited.
'''


import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import uuid

import PlainNet
from PlainNet.net_struct_utils import _get_right_parentheses_index_, create_netblock_list_from_str
from PlainNet.super_blocks import PlainNetSuperBlockClass
from torch import nn
from PlainNet import global_utils


class SuperResIDWEXKX(PlainNetSuperBlockClass):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None,
                 kernel_size=None, expension=None,
                 no_create=False, no_reslink=False, no_BN=False, use_se=False, **kwargs):
        super(SuperResIDWEXKX, self).__init__(**kwargs)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride
        self.bottleneck_channels = bottleneck_channels
        self.sub_layers = sub_layers
        self.kernel_size = kernel_size
        self.expension = expension
        self.no_create = no_create
        self.no_reslink = no_reslink
        self.no_BN = no_BN

        self.use_se = use_se
        if self.use_se:
            print('---debug use_se in ' + str(self))

        full_str = ''
        last_channels = in_channels
        current_stride = stride
        for i in range(self.sub_layers):
            inner_str = ''
            # first DW
            dw_channels = global_utils.smart_round(self.bottleneck_channels * self.expension, base=8)
            inner_str += 'ConvKX({},{},{},{})'.format(last_channels, dw_channels, 1, 1)
            if not self.no_BN:
                inner_str += 'BN({})'.format(dw_channels)
            inner_str += 'RELU({})'.format(dw_channels)

            inner_str += 'ConvDW({},{},{})'.format(dw_channels, self.kernel_size, current_stride)
            if not self.no_BN:
                inner_str += 'BN({})'.format(dw_channels)
            inner_str += 'RELU({})'.format(dw_channels)
            if self.use_se:
                inner_str += 'SE({})'.format(dw_channels)

            inner_str += 'ConvKX({},{},{},{})'.format(dw_channels, bottleneck_channels, 1, 1)
            if not self.no_BN:
                inner_str += 'BN({})'.format(bottleneck_channels)
            # inner_str += 'RELU({})'.format(bottleneck_channels)

            if not self.no_reslink:
                if i == 0:
                    res_str = 'ResBlockProj({})RELU({})'.format(inner_str, self.out_channels)
                else:
                    res_str = 'ResBlock({})RELU({})'.format(inner_str, self.out_channels)

            else:
                res_str = '{}RELU({})'.format(inner_str, self.out_channels)

            full_str += res_str

            # second DW
            inner_str = ''
            dw_channels = global_utils.smart_round(self.out_channels * self.expension, base=8)
            inner_str += 'ConvKX({},{},{},{})'.format(bottleneck_channels, dw_channels, 1, 1)
            if not self.no_BN:
                inner_str += 'BN({})'.format(dw_channels)
            inner_str += 'RELU({})'.format(dw_channels)

            inner_str += 'ConvDW({},{},{})'.format(dw_channels, self.kernel_size, 1)
            if not self.no_BN:
                inner_str += 'BN({})'.format(dw_channels)
            inner_str += 'RELU({})'.format(dw_channels)
            if self.use_se:
                inner_str += 'SE({})'.format(dw_channels)

            inner_str += 'ConvKX({},{},{},{})'.format(dw_channels, self.out_channels, 1, 1)
            if not self.no_BN:
                inner_str += 'BN({})'.format(self.out_channels)

            if not self.no_reslink:
                res_str = 'ResBlock({})RELU({})'.format(inner_str, self.out_channels)
            else:
                res_str = '{}RELU({})'.format(inner_str, self.out_channels)

            full_str += res_str
            last_channels = out_channels
            current_stride = 1
        pass

        self.block_list = create_netblock_list_from_str(full_str, PlainNet._all_netblocks_dict_, no_create=no_create, no_reslink=no_reslink, no_BN=no_BN, **kwargs)
        if not no_create:
            self.module_list = nn.ModuleList(self.block_list)
        else:
            self.module_list = None

    def __str__(self):
        return type(self).__name__ + '({},{},{},{},{})'.format(self.in_channels, self.out_channels,
                                                                self.stride, self.bottleneck_channels, self.sub_layers)

    def __repr__(self):
        return type(self).__name__ + '({}|in={},out={},stride={},btl_channels={},sub_layers={},kernel_size={})'.format(
            self.block_name, self.in_channels, self.out_channels, self.stride, self.bottleneck_channels, self.sub_layers, self.kernel_size
        )

    def encode_structure(self):
        return [self.out_channels, self.sub_layers, self.bottleneck_channels]

    def split(self, split_layer_threshold):
        if self.sub_layers >= split_layer_threshold:
            new_sublayers_1 = split_layer_threshold // 2
            new_sublayers_2 = self.sub_layers - new_sublayers_1
            new_block_str1 = type(self).__name__ + '({},{},{},{},{})'.format(self.in_channels, self.out_channels,
                                                                self.stride, self.bottleneck_channels, new_sublayers_1)
            new_block_str2 = type(self).__name__ + '({},{},{},{},{})'.format(self.out_channels, self.out_channels,
                                                                             1, self.bottleneck_channels,
                                                                             new_sublayers_2)
            return new_block_str1 + new_block_str2
        else:
            return str(self)

    def structure_scale(self, scale=1.0, channel_scale=None, sub_layer_scale=None):
        if channel_scale is None:
            channel_scale = scale
        if sub_layer_scale is None:
            sub_layer_scale = scale

        new_out_channels = global_utils.smart_round(self.out_channels * channel_scale)
        new_bottleneck_channels = global_utils.smart_round(self.bottleneck_channels * channel_scale)
        new_sub_layers = max(1, round(self.sub_layers * sub_layer_scale))

        return type(self).__name__ + '({},{},{},{},{})'.format(self.in_channels, new_out_channels,
                                                                self.stride, new_bottleneck_channels, new_sub_layers)


    @classmethod
    def create_from_str(cls, s, **kwargs):
        assert cls.is_instance_from_str(s)
        idx = _get_right_parentheses_index_(s)
        assert idx is not None
        param_str = s[len(cls.__name__ + '('):idx]

        # find block_name
        tmp_idx = param_str.find('|')
        if tmp_idx < 0:
            tmp_block_name = 'uuid{}'.format(uuid.uuid4().hex)
        else:
            tmp_block_name = param_str[0:tmp_idx]
            param_str = param_str[tmp_idx + 1:]

        param_str_split = param_str.split(',')
        in_channels = int(param_str_split[0])
        out_channels = int(param_str_split[1])
        stride = int(param_str_split[2])
        bottleneck_channels = int(param_str_split[3])
        sub_layers = int(param_str_split[4])
        return cls(in_channels=in_channels, out_channels=out_channels, stride=stride,
                   bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                   block_name=tmp_block_name, **kwargs),s[idx + 1:]


class SuperResIDWE1K3(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE1K3, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=3, expension=1.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE2K3(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE2K3, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=3, expension=2.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE4K3(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE4K3, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=3, expension=4.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE6K3(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE6K3, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=3, expension=6.0,
                                           no_create=no_create, **kwargs)


class SuperResIDWE1K5(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE1K5, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=5, expension=1.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE2K5(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE2K5, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=5, expension=2.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE4K5(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE4K5, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=5, expension=4.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE6K5(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE6K5, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=5, expension=6.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE1K7(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE1K7, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=7, expension=1.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE2K7(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE2K7, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=7, expension=2.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE4K7(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE4K7, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=7, expension=4.0,
                                           no_create=no_create, **kwargs)

class SuperResIDWE6K7(SuperResIDWEXKX):
    def __init__(self, in_channels=None, out_channels=None, stride=None, bottleneck_channels=None, sub_layers=None, no_create=False, **kwargs):
        super(SuperResIDWE6K7, self).__init__(in_channels=in_channels, out_channels=out_channels, stride=stride,
                                           bottleneck_channels=bottleneck_channels, sub_layers=sub_layers,
                                           kernel_size=7, expension=6.0,
                                           no_create=no_create, **kwargs)

def register_netblocks_dict(netblocks_dict: dict):
    this_py_file_netblocks_dict = {
        'SuperResIDWE1K3': SuperResIDWE1K3,
        'SuperResIDWE2K3': SuperResIDWE2K3,
        'SuperResIDWE4K3': SuperResIDWE4K3,
        'SuperResIDWE6K3': SuperResIDWE6K3,
        'SuperResIDWE1K5': SuperResIDWE1K5,
        'SuperResIDWE2K5': SuperResIDWE2K5,
        'SuperResIDWE4K5': SuperResIDWE4K5,
        'SuperResIDWE6K5': SuperResIDWE6K5,
        'SuperResIDWE1K7': SuperResIDWE1K7,
        'SuperResIDWE2K7': SuperResIDWE2K7,
        'SuperResIDWE4K7': SuperResIDWE4K7,
        'SuperResIDWE6K7': SuperResIDWE6K7,
    }
    netblocks_dict.update(this_py_file_netblocks_dict)
    return netblocks_dict


# coding=utf-8
# Copyright (c) 2022, Intel Corporation

# MIT License
# Copyright (c) Microsoft Corporation.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

# =======================================================================================
# MIT license
# =======================================================================================
# - [Swin Transformer](https://github.com/microsoft/swin-transformer)
# - [CLIP](https://github.com/openai/CLIP)

# =======================================================================================
# Apache license 2.0
# =======================================================================================
# - [LeViT](https://github.com/facebookresearch/LeViT)
# - [pytorch-image-models](https://github.com/rwightman/pytorch-image-models)

# =======================================================================================
# BSD-style license
# =======================================================================================
# - [PyTorch](https://github.com/pytorch/pytorch)

import os
import multiprocessing
import tempfile

class _Writer:
    def __init__(self, path, rank):
        self.msg_queue = multiprocessing.Queue()
        self.worker = multiprocessing.Process(
            target=self._async_manager_worker_fn,
            args=(self.msg_queue, path, rank),
        )
        self.worker.start()

    def write(self, key: str, value: bytes) -> bool:
        self.msg_queue.put((key, value))
        return True

    class _WORKER_MSG:
        KILL = 4

    def _async_manager_worker_fn(self, msg_queue, path, rank):
        # path: xxx/logits_top100_epoch0
        rank_name = f'rank{rank}'
        # logits_top100_epoch0_rank0
        basename = os.path.basename(path) + f'_{rank_name}'
        tmp_handle = tempfile.TemporaryDirectory(prefix='tinyvit_' + basename)

        # tmp_dir/tinyvit_logits_top100_epoch0_rank0
        temp_dirname = tmp_handle.name

        tmp_filename = os.path.join(temp_dirname, rank_name)
        # tmp_dir/tinyvit_logits_top100_epoch0_rank0/rank0-keys.txt
        keys_fname = tmp_filename + '-keys.txt'
        values_fname = tmp_filename + '-values.bin'
        keys_file = open(keys_fname, 'w')
        values_file = open(values_fname, 'wb')
        keys = dict()

        try:
            while 1:
                item = msg_queue.get()
                if item == _Writer._WORKER_MSG.KILL:
                    break
                key, value = item
                if key in keys:
                    continue
                idx = len(keys)
                keys[key] = idx
                keys_file.write(key + '\n')
                values_file.write(value)
        except Exception as e:
            print(e)
        finally:
            if hasattr(keys_file, "close"):
                keys_file.close()
            if hasattr(values_file, "close"):
                values_file.close()

        os.makedirs(path, exist_ok=True)
        os.system(f'mv {temp_dirname}/* {path}/')
        print(f"Save logits over: {path}")

    def __del__(self):
        if self.worker is not None:
            self.msg_queue.put(_Writer._WORKER_MSG.KILL)
            self.worker.join()


class _Reader:
    def __init__(self, path: str, item_size: int, rank: int):
        self.rank = rank
        self.item_size = item_size
        self.packages = self.search_packages(path)
        self.packages_visited = [False] * len(self.packages)
        # key -> package idx
        self.keys = dict()

    def read(self, key: str) -> bytes:
        pkg_idx, value_idx = self.keys.get(key, (None, None))
        if pkg_idx is None:
            pkg_idx, value_idx = self.find_item_in_packages(key)
        return self.packages[pkg_idx][value_idx]

    def find_item_in_packages(self, key: str) -> (int, int):
        for pkg_idx, pkg in enumerate(self.packages):
            if not self.packages_visited[pkg_idx]:
                self.packages_visited[pkg_idx] = True
                # load keys
                keys_fname = pkg.name + '-keys.txt'
                with open(keys_fname, 'r') as keys_file:
                    for i, k in enumerate(keys_file.readlines()):
                        k = k.strip()
                        self.keys[k] = (pkg_idx, i)
                if key in self.keys:
                    return self.keys[key]
        raise KeyError(key)

    def search_packages(self, path):
        names = self.search_packages_names(path)
        return [_Reader._PackageReader(name, self.item_size) for name in names]

    def search_packages_names(self, path):
        names = []
        VALUES_POSTFIX = '-values.bin'
        for name in os.listdir(path):
            if name.endswith(VALUES_POSTFIX):
                names.append(name[:-len(VALUES_POSTFIX)])

        num_packages = len(names)

        def rank_key_fn(name):
            r = int(name[4:])
            return (r - self.rank) % num_packages

        # move the rankx-keys.txt to the front
        names.sort(key=rank_key_fn)
        names = list(map(lambda x: os.path.join(path, x), names))
        return names

    class _PackageReader:
        def __init__(self, name, item_size):
            self.name = name
            self.item_size = item_size

            # delay to create handle
            self.values_file = None

        def __getitem__(self, idx: int):
            self._ensure_handle_created()
            self.values_file.seek(self.item_size * idx)
            try:
                values_content = self.values_file.read(self.item_size)
            except Exception as e:
                print(e)
            # finally:
            #     if hasattr(self.values_file, 'close'):
            #         self.values_file.close()
            return values_content

        def _ensure_handle_created(self):
            if self.values_file is None:
                values_fname = self.name + '-values.bin'
                self.values_file = open(values_fname, 'rb')


class TxtManager:
    def __init__(self, path: str, item_size: int, rank: int):
        self.path = path
        self.writer = None
        self.reader = None
        self.item_size = item_size
        self.rank = rank

    def write(self, key: str, value: bytes) -> bool:
        if self.writer is None:
            self.writer = _Writer(self.path, self.rank)
        return self.writer.write(key, value)

    def read(self, key: str) -> bytes:
        if self.reader is None:
            self.reader = _Reader(self.path, self.item_size, self.rank)
        return self.reader.read(key)



import torch
from e2eAIOK.common.trainer.data_builder import DataBuilder
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist
from e2eAIOK.DeNas.asr.data.dataio.dataset import dataio_prepare
from e2eAIOK.DeNas.asr.data.dataio.batch import PaddedBatch

class DataBuilderASR(DataBuilder):
    def __init__(self, cfg, tokenizer):
        super().__init__(cfg)
        self.tokenizer = tokenizer

    def get_dataloader(self):
        if self.dataset_train is None or self.dataset_val is None:
            self.prepare_dataset()

        if ext_dist.my_size > 1:
            num_tasks = ext_dist.dist.get_world_size()
            global_rank = ext_dist.dist.get_rank()
            sampler_train = torch.utils.data.DistributedSampler(
                self.dataset_train, num_replicas=num_tasks, rank=global_rank
            )
            
            sampler_val = torch.utils.data.DistributedSampler(
                self.dataset_val, num_replicas=num_tasks, rank=global_rank)
        else:
            sampler_val = None
            sampler_train = None

        dataloader_train = torch.utils.data.DataLoader(
            self.dataset_train, 
            sampler=sampler_train,
            batch_size=self.cfg.train_batch_size,
            num_workers=self.cfg.num_workers,
            collate_fn=PaddedBatch,
            shuffle=(ext_dist.my_size <= 1),
            drop_last=True,
        )

        dataloader_val = torch.utils.data.DataLoader(
            self.dataset_val, 
            batch_size=self.cfg.eval_batch_size,
            sampler=sampler_val, 
            num_workers=self.cfg.num_workers,
            collate_fn=PaddedBatch,
            shuffle=False,
            drop_last=False
        )
        
        return dataloader_train, dataloader_val


import os
import json
from collections import OrderedDict


class HydroModel:
    """
    A HydroModel instance contains infomation of one trained model

    Attributes
    ----------
    model_params : dict
        parameters for this model, include model type, global settings
        and sigopt parameters
    model : str
        model saved path
    metrics : list
        metrics used to communicate with sigopt for finding best model,
        maybe 'AUC', 'MAP', 'training_time', etc
    sigopt_experiment_id: str
        If this model is trained with sigopt, record its history sigopt
        experiment id
    model_parameters: dict
        model parameters used to train this model, can be parameter
        suggested by sigopt or user defined.
    """
    def __new__(cls, settings=None, serialized_text=None, hydro_model_path=""):
        if hydro_model_path != "" and not os.path.exists(hydro_model_path):
            return None
        obj = super().__new__(cls)
        obj.__init__(settings, serialized_text, hydro_model_path)
        return obj

    def __init__(self, settings=None, serialized_text=None, hydro_model_path=""):
        """
        Create HydroModel instance

        Instance can be created either from settings or a in database
        json string

        Parameters
        ----------
        settings : dict
            optional
        serialized_text : str
            Json string, optional
        """ 
        if hydro_model_path != "" and os.path.exists(hydro_model_path):
            with open(f"{hydro_model_path}", 'r') as f:
                jdata = f.read()
            serialized_text=[jdata]
        if serialized_text:
            self.load_json(serialized_text)
        else:
            self.model_params = settings
            self.model = ""
            self.metrics = []
            self.sigopt_experiment_id = None
            self.model_parameters = None

    def update(self, info={}):
        """
        This method is used to update current HydroModel

        Parameters
        ----------
        info: dict:
            example info as {'model': 'saved_model_path', 'metrics':
            ['AUC': 8.2127], 'sigopt_experiment_id': 'xxxx',
            'model_parameters': {'learning_rt': 3}}
        """
        for k, v in info.items():
            if k == 'model':
                self.model = v
            elif k == 'metrics':
                self.metrics = v
            elif k == 'sigopt_experiment_id':
                self.sigopt_experiment_id = v
            elif k == 'model_parameters':
                self.model_parameters = v

    def load_json(self, serialized_text):
        """
        Load serialized text to HydroModel

        We will use Json format to store HydroModel in HydroDB, this
        method is used to deserialize Json

        Parameters
        ----------
        serialized_text : str
        """
        data = json.loads(serialized_text[0])
        self.model = data['model']
        self.metrics = data['metrics']
        self.sigopt_experiment_id = data['sigopt_experiment_id']
        self.model_params = data['model_params']
        self.model_parameters = data['model_parameters']

    def to_json(self):
        """
        serialize HydroModel to json

        We will use Json format to store HydroModel in HydroDB, this
        method is used to serialize HydroModel

        Returns
        ----------
        out : str
        """
        data = OrderedDict(sorted(self.__dict__.items()))
        out = json.dumps(data)
        return out

    def explain(self):
        """
        Explain this model
        """
        print("\n===============================================")
        print(f"***    Best Trained Model    ***")
        print("===============================================")
        print(f"  Model Type: {self.model_params['model_name']}")
        print(f"  Model Saved Path: {self.model}")
        print(f"  Sigopt Experiment id is {self.sigopt_experiment_id}")
        print("  === Result Metrics ===")
        for item in self.metrics:
            print(f"    {item['name']}: {item['value']}")
        print("===============================================")

    def save_to_path(self, save_path):
        with open(f"{save_path}", 'w') as f:
            f.write(self.to_json())


'''
Copyright (C) 2010-2021 Alibaba Group Holding Limited.
'''

import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import torch, argparse
from torch import nn
from PlainNet.net_struct_utils import _create_netblock_list_from_str_

_all_netblocks_dict_ = {}

class PlainNet(nn.Module):
    def __init__(self, num_classes=None, plainnet_struct=None, no_create=False, **kwargs):
        super(PlainNet, self).__init__()
        self.num_classes = num_classes
        self.plainnet_struct = plainnet_struct

        if self.plainnet_struct is None:
            raise ValueError("plainnet_struct is None")

        if self.plainnet_struct is None:
            return

        the_s = self.plainnet_struct  # type: str

        block_list, remaining_s = _create_netblock_list_from_str_(the_s, _all_netblocks_dict_,  no_create=no_create, **kwargs)
        assert len(remaining_s) == 0

        self.block_list = block_list
        if not no_create:
            self.module_list = nn.ModuleList(block_list)  # register

    def forward(self, x):
        output = x
        for the_block in self.block_list:
            output = the_block(output)
        return output

    def __str__(self):
        s = ''
        for the_block in self.block_list:
            s += str(the_block)
        return s

    def __repr__(self):
        return str(self)

    @classmethod
    def create_netblock_list_from_str(cls, s, no_create=False, **kwargs):
        the_list, remaining_s = _create_netblock_list_from_str_(s, _all_netblocks_dict_, no_create=no_create, **kwargs)
        assert len(remaining_s) == 0
        return the_list

    def get_FLOPs(self, input_resolution):
        the_res = input_resolution
        the_flops = 0
        for the_block in self.block_list:
            the_flops += the_block.get_FLOPs(the_res)
            the_res = the_block.get_output_resolution(the_res)

        return the_flops

    def get_model_size(self):
        the_size = 0
        for the_block in self.block_list:
            the_size += the_block.get_model_size()

        return the_size

    def replace_block(self, block_id, new_block):
        self.block_list[block_id] = new_block
        if block_id < len(self.block_list):
            self.block_list[block_id + 1].set_in_channels(new_block.out_channels)

        self.module_list = nn.Module(self.block_list)



from PlainNet import basic_blocks
_all_netblocks_dict_ = basic_blocks.register_netblocks_dict(_all_netblocks_dict_)

from PlainNet import super_blocks
_all_netblocks_dict_ = super_blocks.register_netblocks_dict(_all_netblocks_dict_)

from PlainNet import SuperResKXKX
_all_netblocks_dict_ = SuperResKXKX.register_netblocks_dict(_all_netblocks_dict_)

from PlainNet import SuperResK1KXK1
_all_netblocks_dict_ = SuperResK1KXK1.register_netblocks_dict(_all_netblocks_dict_)

from PlainNet import SuperResIDWEXKX
_all_netblocks_dict_ = SuperResIDWEXKX.register_netblocks_dict(_all_netblocks_dict_)


# coding=utf-8
# Copyright (c) 2022, Intel Corporation

# MIT License
# Copyright (c) Microsoft Corporation.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

# =======================================================================================
# MIT license
# =======================================================================================
# - [Swin Transformer](https://github.com/microsoft/swin-transformer)
# - [CLIP](https://github.com/openai/CLIP)

# =======================================================================================
# Apache license 2.0
# =======================================================================================
# - [LeViT](https://github.com/facebookresearch/LeViT)
# - [pytorch-image-models](https://github.com/rwightman/pytorch-image-models)

# =======================================================================================
# BSD-style license
# =======================================================================================
# - [PyTorch](https://github.com/pytorch/pytorch)

import os
import multiprocessing
import torch
import torch.distributed as dist
import numpy as np
from .aug_random import AugRandomContext
from .manager import TxtManager


def get_rank():
    if not dist.is_available() or not dist.is_initialized():
        return 0
    return dist.get_rank()


class DatasetWrapper(torch.utils.data.Dataset):
    def __init__(self, dataset, logits_path, num_classes, topk, write):
        super().__init__()
        self.dataset = dataset
        self.logits_path = logits_path
        self.epoch = multiprocessing.Value('i', 0)
        self.topk = topk
        self.num_classes = num_classes
        self.write_mode = write
        self.keys = self._get_keys()
        self._manager = (None, None)

    def __getitem__(self, index: int):
        if self.write_mode:
            return self.__getitem_for_write(index)
        return self.__getitem_for_read(index)

    def __getitem_for_write(self, index: int):
        # get an augmentation seed
        key = self.keys[index]
        seed = np.int32(np.random.randint(0, 1 << 31))
        with AugRandomContext(seed=int(seed)):
            item = self.dataset[index]
        return (item, (key, seed))

    def __getitem_for_read(self, index: int):
        key = self.keys[index]
        values = self._get_saved_logits(key)
        seed = values[0]
        with AugRandomContext(seed=seed):
            item = self.dataset[index]

        if self.topk == 0:
            return ((item[0], (values[1], np.int32(seed))), item[1])
        elif self.topk > 0:
            return ((item[0], (values[1], values[2], np.int32(seed))), item[1])
        
    def _get_saved_logits(self, key: str):
        manager = self.get_manager()
        bstr: bytes = manager.read(key)
        # parse the augmentation seed
        seed = int(np.frombuffer(bstr[:4], dtype=np.int32))
        # parse the logits index and value
        # copy logits_index and logits_value to avoid warning of written flag from PyTorch
        bstr = bstr[4:]
        if self.topk == 0:
            logits_value = np.frombuffer(
                bstr[:self.num_classes * 4], dtype=np.float32).copy()
            return seed, logits_value
        elif self.topk > 0:
            logits_value = np.frombuffer(
                bstr[:self.topk * 2], dtype=np.float16).copy()
            bstr = bstr[self.topk * 2:]
            logits_index = np.frombuffer(
                bstr[:self.topk * 2], dtype=np.int16).copy()
            return seed, logits_value, logits_index

    def _build_manager(self, logits_path: str):
        if self.topk == 0: # save all logits
            # num_classes * 4 bytes for logits + 4 bytes for seed
            item_size = self.num_classes * 4 + 4
        elif self.topk > 0: # save topk logits
            # topk * [idx, value] * 2 bytes  for logits + 4 bytes for seed
            item_size = self.topk * 2 * 2 + 4
        rank = get_rank()
        return TxtManager(logits_path, item_size, rank)

    def set_epoch(self, epoch: int):
        self.epoch.value = epoch
        self._manager = (None, None)

    def get_manager(self):
        epoch = self.epoch.value
        if epoch != self._manager[0]:
            if self.topk ==0 :
                logits_path = os.path.join(
                    self.logits_path, f'logits_epoch{self.epoch.value}')
            elif self.topk > 0:
                logits_path = os.path.join(
                    self.logits_path, f'logits_top{self.topk}_epoch{self.epoch.value}')
            self._manager = (epoch, self._build_manager(logits_path))
        return self._manager[1]

    def __len__(self):
        return len(self.dataset)

    def _get_keys(self):
        if hasattr(self.dataset, 'get_keys'):
            keys = self.dataset.get_keys()
            if self.write_mode:
                # we only check key unique in the write mode
                assert len(keys) == len(set(keys)), 'keys must be unique'
            return keys
        return [str(i) for i in range(len(self))]



from easydict import EasyDict as edict
import os, time, datetime
import yaml
import torch
import torch.nn as nn
import logging
import torchvision
import optuna
from optuna.trial import TrialState
import argparse
from functools import partial
import sys
import e2eAIOK
from e2eAIOK.ModelAdapter.training import ModelAdapterTask
from e2eAIOK.common.utils import update_dict
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist
e2eaiok_dir = e2eAIOK.__path__[0]

def is_safe_path(basedir, path):
    return os.path.abspath(path).startswith(basedir)

def objective(args,trial):
    ''' Optuna optimize objective
    Args:
        args: parsed args
        trial: a trial object
    Returns: a metric to measure the performance of model
    '''
    if "MASTER_ADDR" not in os.environ:
        os.environ["MASTER_ADDR"] = "127.0.0.1"
    if "MASTER_PORT" not in os.environ:
        os.environ["MASTER_PORT"] = "8089"
    return main(args, trial)

def main(args, trial):
    ''' main function
    :param args: args parameters.
    :param trial: Optuna Trial. If None, then training without automl.
    :return: validation metric. If has Earlystopping, using the best metric; Else, using the last metric.
    '''
    #################### merge cfg ################
    with open(os.path.join(e2eaiok_dir, "common/default.conf")) as f:
        cfg = yaml.safe_load(f)
    with open(os.path.join(e2eaiok_dir, "ModelAdapter/default_ma.conf")) as f:
        cfg = update_dict(cfg, yaml.safe_load(f))
    if not is_safe_path("/home/vmagent/app/e2eaiok", args.cfg):
        print(f"{args.cfg} is not safe.")
        sys.exit()      
    if os.access(args.cfg, os.R_OK):
        with open(args.cfg) as f:
            cfg = edict(update_dict(cfg, yaml.safe_load(f)))
    with open(args.cfg) as f:
        cfg = edict(update_dict(cfg, yaml.safe_load(f)))
    torch.manual_seed(cfg.seed)

    ## To be fix
    if cfg.num_workers > 0:
        raise RuntimeError("Do not support dataloader number of workers > 0, will be fixed later.")
        sys.exit(0)

    #################### dir conguration ################
    is_distributed = ext_dist.my_size > 1
    prefix = "%s%s%s%s%s"%(cfg.model_type,
                     "_%s"%cfg.experiment.strategy if cfg.experiment.strategy else "",
                     "_%s"%cfg.data_set,
                     "_trial%s" % trial.number if trial is not None else "",
                     "_rank%s" % ext_dist.my_rank if is_distributed else "")
    prefix_time = "%s_%s"%(prefix,int(time.time()))
    cfg.experiment.tag = cfg.experiment.tag + "%s" % ("_dist%s" % ext_dist.my_size if is_distributed else "")
    root_dir = os.path.join(cfg.output_dir, cfg.experiment.project,cfg.experiment.tag)
    LOG_DIR = os.path.join(root_dir,"log")                      # to save training log
    PROFILE_DIR = os.path.join(root_dir,"profile")              # to save profiling result
    model_save_path = os.path.join(root_dir, prefix)
    cfg.tensorboard_dir = os.path.join(cfg.tensorboard_dir,"%s_%s"%(cfg.experiment.tag,prefix))  # to save tensorboard log
    cfg.profiler_config.trace_file = os.path.join(PROFILE_DIR,"profile_%s"%prefix_time)
    if not is_safe_path("/home/vmagent/app/data", LOG_DIR):
        print(f"{LOG_DIR} is not safe.")
        sys.exit()
    if not is_safe_path("/home/vmagent/app/data", PROFILE_DIR):
        print(f"{PROFILE_DIR} is not safe.")
        sys.exit()
    if not is_safe_path("/home/vmagent/app/data", cfg.tensorboard_dir):
        print(f"{cfg.tensorboard_dir} is not safe.")
        sys.exit()
    if not is_safe_path("/home/vmagent/app/data", model_save_path):
        print(f"{model_save_path} is not safe.")
        sys.exit()    
    os.makedirs(LOG_DIR,exist_ok=True)
    os.makedirs(PROFILE_DIR,exist_ok=True) 
    os.makedirs(cfg.tensorboard_dir,exist_ok=True)
    os.makedirs(model_save_path,exist_ok=True)

    if int(cfg.distiller.save_logits) + int(cfg.distiller.use_saved_logits) + int(cfg.distiller.check_logits) >=2:
        raise RuntimeError("Can not save teacher logits, train students with logits or check logits together!")
    if cfg.distiller.save_logits:
        os.makedirs(cfg.distiller.logits_path, exist_ok=True)
    if cfg.distiller.use_saved_logits or cfg.distiller.check_logits:
        if not os.path.exists(cfg.distiller.logits_path):
            raise RuntimeError("Need teacher saved logits!")
 
    #################### logging conguration ################
    # Remove all handlers associated with the root logger object.
    for handler in logging.root.handlers[:]: 
        logging.root.removeHandler(handler)
    
    log_filename = os.path.join(LOG_DIR, "%s.txt"%prefix_time)
    logging.basicConfig(filename=log_filename, level=logging.INFO,
                        format='%(asctime)s %(levelname)s [%(filename)s %(funcName)s %(lineno)d]: %(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S %p',
                        filemode='w')
    ################ init dist ################
    ext_dist.init_distributed(backend=cfg.dist_backend)
    ##################### Optuna hyper params ################
    if trial is not None:
        cfg.learning_rate = trial.suggest_float("lr", 0.001, 0.1, log=True)
        cfg.finetuner.learning_rate = trial.suggest_float("lr_finetuned", 0.001, 0.1, log=True) 
        cfg.weight_decay = trial.suggest_float("weight_decay", 0.0001, 0.1,log=True)
    ##################### show conguration ################
    print("configurations:")
    print(cfg)
    ###################### create task ###############
    task = ModelAdapterTask(cfg, model_save_path, is_distributed)
    if trial is not None:
        logging.info("[%s]: Begin trial %s" % (datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), trial.number))
        print("[%s]: Begin trial %s" % (datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), trial.number))
    metric = task.run(eval=args.eval, resume=args.resume)
    ############### destroy dist ###############
    if is_distributed:
        dist.destroy_process_group()
    if trial is not None:
        logging.info("End trial %s" % trial.number)
        print("[%s]: End trial %s" % (datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), trial.number))
    return metric

if __name__ == '__main__':
    # usage: python main.py --cfg ../config/demo/cifar100_kd_vit_res18.yaml --opts train_epochs 1 dataset.path /xxx/yyy
    start_time = time.time()
    parser = argparse.ArgumentParser()
    parser.add_argument("--cfg", type=str, default="../config/baseline/cifar100_resnet18_LRdecay10.conf")
    parser.add_argument('--eval',action='store_true')
    parser.add_argument('--resume',action='store_true')
    parser.add_argument('-R', "--trial_round", default=0,help="The hyper-param tunning round. trial_round <= 0 means no tunning.", type=int)
    # parser.add_argument("--opts", default=None, nargs=argparse.REMAINDER)
    args = parser.parse_args()

    trial_round = parser.parse_args().trial_round
    if trial_round > 0:
        study = optuna.create_study(direction="maximize")
        study.optimize(partial(objective,args), n_trials=trial_round, timeout=None)

        pruned_trials = study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
        complete_trials = study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])

        print("Study statistics: ")
        print("  Number of finished trials: ", len(study.trials))
        print("  Number of pruned trials: ", len(pruned_trials))
        print("  Number of complete trials: ", len(complete_trials))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: ", trial.value)

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))
    else:
        print("No Trial")
        objective(args,None)

    print(f"Totally take {(time.time()-start_time)} seconds")


import sys

from e2eAIOK.DeNas.asr.utils import edit_distance


def print_wer_summary(wer_details, file=sys.stdout):
    print(
        "%WER {WER:.2f} [ {num_edits} / {num_scored_tokens}, {insertions} ins, {deletions} del, {substitutions} sub ]".format(  # noqa
            **wer_details
        ),
        file=file,
        end="",
    )
    print(
        " [PARTIAL]"
        if wer_details["num_scored_sents"] < wer_details["num_ref_sents"]
        else "",
        file=file,
    )
    print(
        "%SER {SER:.2f} [ {num_erraneous_sents} / {num_scored_sents} ]".format(
            **wer_details
        ),
        file=file,
    )
    print(
        "Scored {num_scored_sents} sentences, {num_absent_sents} not present in hyp.".format(  # noqa
            **wer_details
        ),
        file=file,
    )


def print_alignments(
    details_by_utterance,
    file=sys.stdout,
    empty_symbol="<eps>",
    separator=" ; ",
    print_header=True,
    sample_separator=None,
):
    if print_header:
        _print_alignments_global_header(
            file=file, empty_symbol=empty_symbol, separator=separator
        )
    for dets in details_by_utterance:
        if dets["scored"]:
            if print_header:
                _print_alignment_header(dets, file=file)
            _print_alignment(
                dets["alignment"],
                dets["ref_tokens"],
                dets["hyp_tokens"],
                file=file,
                empty_symbol=empty_symbol,
                separator=separator,
            )
            if sample_separator:
                print(sample_separator, file=file)


def _print_alignment(
    alignment, a, b, empty_symbol="<eps>", separator=" ; ", file=sys.stdout
):
    # First, get equal length text for all:
    a_padded = []
    b_padded = []
    ops_padded = []
    for op, i, j in alignment:  # i indexes a, j indexes b
        op_string = str(op)
        a_string = str(a[i]) if i is not None else empty_symbol
        b_string = str(b[j]) if j is not None else empty_symbol
        # NOTE: the padding does not actually compute printed length,
        # but hopefully we can assume that printed length is
        # at most the str len
        pad_length = max(len(op_string), len(a_string), len(b_string))
        a_padded.append(a_string.center(pad_length))
        b_padded.append(b_string.center(pad_length))
        ops_padded.append(op_string.center(pad_length))
    # Then print, in the order Ref, op, Hyp
    print(separator.join(a_padded), file=file)
    print(separator.join(ops_padded), file=file)
    print(separator.join(b_padded), file=file)


def _print_alignments_global_header(
    empty_symbol="<eps>", separator=" ; ", file=sys.stdout
):
    print("=" * 80, file=file)
    print("ALIGNMENTS", file=file)
    print("", file=file)
    print("Format:", file=file)
    print("<utterance-id>, WER DETAILS", file=file)
    # Print the format with the actual
    # print_alignment function, using artificial data:
    a = ["reference", "on", "the", "first", "line"]
    b = ["and", "hypothesis", "on", "the", "third"]
    alignment = [
        (edit_distance.EDIT_SYMBOLS["ins"], None, 0),
        (edit_distance.EDIT_SYMBOLS["sub"], 0, 1),
        (edit_distance.EDIT_SYMBOLS["eq"], 1, 2),
        (edit_distance.EDIT_SYMBOLS["eq"], 2, 3),
        (edit_distance.EDIT_SYMBOLS["sub"], 3, 4),
        (edit_distance.EDIT_SYMBOLS["del"], 4, None),
    ]
    _print_alignment(
        alignment,
        a,
        b,
        file=file,
        empty_symbol=empty_symbol,
        separator=separator,
    )


def _print_alignment_header(wer_details, file=sys.stdout):
    print("=" * 80, file=file)
    print(
        "{key}, %WER {WER:.2f} [ {num_edits} / {num_ref_tokens}, {insertions} ins, {deletions} del, {substitutions} sub ]".format(  # noqa
            **wer_details
        ),
        file=file,
    )



import argparse
import pathlib
import sys

import yaml
from e2eAIOK.dataloader.hydrodataloader import *
from e2eAIOK.utils.hydroconfig import *
from e2eAIOK.utils.hydromodel import *

from e2eAIOK.SDA.modeladvisor.DIENAdvisor import *
from e2eAIOK.SDA.modeladvisor.TestAdvisor import *
from e2eAIOK.SDA.modeladvisor.DLRMAdvisor import *
from e2eAIOK.SDA.modeladvisor.WnDAdvisor import *
from e2eAIOK.SDA.modeladvisor.TwitterRecSysAdvisor import *
from e2eAIOK.SDA.modeladvisor.MiniGoAdvisor import *
from e2eAIOK.SDA.modeladvisor.RNNTAdvisor import *
from e2eAIOK.SDA.modeladvisor.BERTAdvisor import *
from e2eAIOK.SDA.modeladvisor.ResNetAdvisor import *
from e2eAIOK.SDA.modeladvisor.UPMAdvisor import *
from e2eAIOK.SDA.modeladvisor.RegisteredAdvisor import *

class SDA:
    """
    Smart Democratization Advisor creates sigopt parameter or model
    parameter based on different model type

    Attributes
    ----------
    model : str
        model type, can be provided models such as dlrm, dien, wnd
        or udm(user defined model)
    data_loader: HydroDataLoader
        A dataset iterator or infomation container
    settings: dict
        passed-in configuration includes global configuration, sigopt
        config and model parameters
    hydro_model: HydroModel
        history best model object, if this attribute is not None, we
        can resume from a history experiment.
    """
    def __init__(self, model="custom_registered", data_loader=None, settings={}, hydro_model=None, custom_result_path=None):
        """
        Parameters
        ----------
        model: str
            input model type, can be in-stock-model or user-defined-model
        data_loader: HydroDataLoader
            dataset loader
        settings: dict
            input parameters includes global configuration and model
            parameter configuration
        hydro_model: HydroModel
            history best model object, this parameter will only be
            used with e2eaiok, optional.
        """
        self.model = model
        if data_loader is None and 'data_path' in settings:
            data_loader = HydroDataLoaderAdvisor.create_data_loader(
                settings['data_path'], self.model)
        self.data_loader = data_loader
        self.custom_result_path = custom_result_path
        if self.data_loader != None:
            self.dataset_meta = self.data_loader.get_meta()
            self.dataset_train = self.data_loader.get_train()
            self.dataset_valid = self.data_loader.get_valid()
        else:
            self.dataset_meta = None
            self.dataset_train = None
            self.dataset_valid = None
        self.settings = default_settings(model, settings)
        self.hydro_model = hydro_model
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('E2EAIOK.SDA')
        self.logger.info("""### Ready to submit current task  ###""")

    def __del__(self):
        with open(f"{self.custom_result_path}/latest_hydro_model", 'w') as f:
            f.write(self.hydro_model.to_json())

    def __create_model_advisor(self):
        if self.model.lower() == 'wnd':
            return WnDAdvisor(self.dataset_meta, self.dataset_train,
                              self.dataset_valid, self.settings)
        elif self.model.lower() == 'dlrm':
            return DLRMAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        elif self.model.lower() == 'dien':
            return DIENAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        elif self.model.lower() == 'twitter_recsys':
            return TwitterRecSysAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        elif self.model.lower() == 'upm':
            return UPMAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        elif self.model.lower() == 'pipeline_test':
            return TestAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        elif self.model.lower() == "minigo":
                return MiniGoAdvisor(self.dataset_meta, self.dataset_train,
                                   self.dataset_valid, self.settings)
        elif self.model.lower() == 'rnnt':
            return RNNTAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        elif self.model.lower() == 'bert':
            return BERTAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        elif self.model.lower() == 'resnet':
            return ResNetAdvisor(self.dataset_meta, self.dataset_train,
                               self.dataset_valid, self.settings)
        else:
            return RegisteredAdvisor(settings=self.settings)

    @staticmethod
    def get_model_zoo_list():
        return ['wnd', 'dlrm', 'dien', 'twitter_recsys', 'rnnt', 'minigo', 'bert', 'resnet', 'pipeline_test', 'upm']

    def launch(self):
        """
        Launch a new SDA task

        This method will generate a SDA instance to generate model
        parameters based on model type and user input, then iterating
        train and sigopt communication to find the best model

        Returns
        -------
        model_path: str
            best trained model saved path
        metrics: list
            A list of this best trained model infomation
        """
        # 1. get model advisor
        # sigopt yaml will be created and sigopt connection will be setup
        if "model_advisor" not in dir(self):
            self.model_advisor = self.__create_model_advisor()
            self.logger.info("Model Advisor created")

        # 2. initialize_sigopt
        if self.settings["enable_sigopt"]:
            self.logger.info("Start to init sigopt")
            experiment_id = self.hydro_model.sigopt_experiment_id if self.hydro_model else None
            experiment_id = self.model_advisor.initialize_sigopt(
                experiment_id=experiment_id)
            if not self.hydro_model:
                self.hydro_model = HydroModel(self.settings)
            self.hydro_model.update(
                {'sigopt_experiment_id': experiment_id})
            self.model_advisor.record_assignmet(
                self.hydro_model.model_parameters, self.hydro_model.model)
        else:
            best_model_parameters = self.hydro_model.model_parameters if self.hydro_model else None
            if not self.hydro_model:
                self.hydro_model = HydroModel(self.settings)
            self.model_advisor.initialize_model_parameter(
                assignments=best_model_parameters)
        self.logger.info("model parameter initialized")

        # 3. launch train, w/wo sigopt to iterate train multiple times until we
        # reached numIter or target score
        self.logger.info("start to launch training")
        model_path, metrics, parameters = self.model_advisor.launch_train()
        self.hydro_model.update({
            'model': model_path,
            'metrics': metrics,
            'model_parameters': parameters
        })
        self.logger.info("training script completed")
        return model_path, metrics

    def snapshot(self):
        return self.hydro_model

    def register(self, info):
        if "model_advisor" not in dir(self):
            self.model_advisor = self.__create_model_advisor()
            self.logger.info("Model Advisor created")
        self.model_advisor.register(info)

def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument('--model_name',
        type=str,
        required=True,
        help='could be in-stock model name or udm(user-define-model)')
    parser.add_argument('--data_path',
        type=str,
        required=True,
        help='Dataset path')
    parser.add_argument('--conf',
                        type=str,
                        default='conf/e2eaiok_defaults.conf',
                        help='e2eaiok defaults configuration')
    parser.add_argument('--custom_result_path',
        type=str,
        default=str(pathlib.Path(__file__).parent.absolute()),
        help='custom result path')
    parser.add_argument('--enable_sigopt',
        dest="enable_sigopt",
        action="store_true",
        default=False,
        help='if enable sigopt')
    parser.add_argument('--no_model_cache',
        dest="enable_model_cache",
        action="store_false",
        default=True,
        help='if disable model cache')
    parser.add_argument('--interactive',
        dest="interative",
        action="store_true",
        help='enable interative mode')
    return parser.parse_args(args).__dict__


def main(input_args):
    settings = init_settings()
    settings.update(input_args)
    settings.update(parse_config(settings['conf']))
    data_loader = HydroDataLoaderAdvisor.create_data_loader(
        settings['data_path'], settings['model_name'])
    current_path = settings['custom_result_path']
    if settings["enable_model_cache"] and os.path.exists(f"{current_path}/latest_hydro_model"):
        with open(f"{current_path}/latest_hydro_model", 'r') as f:
            jdata = f.read()
        hydro_model = HydroModel(None, serialized_text=[jdata])
        if hydro_model.model_params['model_name'] == settings['model_name']:
            hydro_model.explain()
            r = timeout_input("Do you want to use this history hydro model? y or n", 'n', 10, interative = settings["interative"])
            if r == 'n':
                print("Skip history hydro model, create new hydro model")
                hydro_model = HydroModel(settings)
            self.logger.info("""Above info is history record of this model""")
        else:
            print("Detected history hydro model, but skip since model type is not the same")
            hydro_model = HydroModel(settings)
    else:
        hydro_model = HydroModel(settings)
    sda = SDA(settings['model_name'], data_loader, settings, hydro_model)
    sda.launch()
    with open(f"{current_path}/latest_hydro_model", 'w') as f:
        f.write(hydro_model.to_json())
    hydro_model.explain()


if __name__ == '__main__':
    input_args = parse_args(sys.argv[1:])
    main(input_args)



import torch
import datetime, time
import contextlib
import os
import numpy as np
from collections.abc import Iterable
from e2eAIOK.common.trainer.torch_trainer import TorchTrainer
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist
from types import MethodType

class TorchTrainerMA(TorchTrainer):
    ''' Trainer for Model Adapter

    '''
    def __init__(self, cfg, model, metric, best_metric_name, tensorboard_writer=None, 
                    train_dataloader=None, eval_dataloader=None, is_transferrable=False,
                    optimizer=None, scheduler=None, warmup_scheduler=None, 
                    criterion=None, early_stopping=None, 
                    profiler=None, device='cpu'):
        ''' Init method
        :param cfg: configurations
        :param model: the trained model
        :param metric: metric function map, which map metric name to metric function
        :param best_metric_name: metric name for update best model
        :param tensorboard_writer: tensorboard writer, can be None
        :param train_dataloader: train dataloader, can be None in evaluate mode
        :param eval_dataloader: validation dataloader
        :param is_transferrable: is model transferrable
        :param optimizer: optimizer
        :param scheduler: learning rate scheduler, can be None
        :param warmup_scheduler: the scheduler for warmup, can be None
        :param criterion: criterion for loss function, can be None
        :param early_stopping: for early stopping, can be None
        :param profiler : training profiler
        :param device: running on cpu or gpu
        '''
        super().__init__(cfg, model, train_dataloader, eval_dataloader, optimizer, criterion, scheduler, metric)
        self.best_metric_name = best_metric_name
        self.tensorboard_writer = tensorboard_writer
        self.warmup_scheduler = warmup_scheduler
        self.early_stopping = early_stopping
        self.is_transferrable = is_transferrable
        self.profiler = profiler
        self.device = device
        self.best_metric_value = 0.0
        self.backbone = self.model.get_backbone() if self.is_transferrable else self.model
        if best_metric_name not in metric:
            raise RuntimeError("early stop metric [%s] not in metric keys [%s]"%(
                best_metric_name,",".join(metric.keys())
            ))

    def __str__(self):
        _str = "Trainer: model:%s\n"%self.model
        _str += "\tconfigurations:%s\n" % self.cfg
        _str += "\ttrain_dataloader:%s\n"%self.train_dataloader
        _str += "\teval_dataloader:%s\n"%self.eval_dataloader
        _str += "\toptimizer:%s\n"%self.optimizer
        _str += "\tscheduler:%s\n" % self.scheduler
        _str += "\tmetric:%s\n" % self.metric
        _str += "\tcriterion:%s\n" % self.criterion
        _str += "\t_best_metric_name:%s\n" % self.best_metric_name
        _str += "\ttensorboard_writer:%s\n" % self.tensorboard_writer
        _str += "\twarmup_scheduler:%s\n" % self.warmup_scheduler
        _str += "\tearly_stopping:%s\n" % self.early_stopping
        _str += "\tis_transferrable:%s\n" % self.is_transferrable
        _str += "\tprofiler:%s\n" % self.profiler
        _str += "\tdevice:%s\n" % self.device
        _str += "\trank:%s\n" % ext_dist.my_rank
        _str += "\tbest_metric_value:%s\n" % self.best_metric_value
        _str += "\ttraining_epochs:%s\n" % self.cfg.train_epochs
        _str += "\tlogging_interval:%s\n" % self.cfg.log_interval_step
        return _str

    def show_update_tensorboard_metric(self, dataset_name,metric_values,cur_epoch=0,cur_step=0,epoch_steps=0):
        ''' add metric to tensorboard

        :param dataset_name: dataset name (Train? Evaluation? Test?)
        :param metric_values: metric name and metric value
        :param cur_epoch: current epoch
        :param cur_step: current step
        :param epoch_steps : step num of one epoch
        :return:
        '''
        if dataset_name not in ['Train','Validation','Test']:
            raise RuntimeError("dataset_name (%s) must in 'Train','Validation','Test'"%dataset_name)

        dt = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S') # date time
        if self.tensorboard_writer is not None:
            for (metric_name, metric_value) in metric_values.items():
                self.tensorboard_writer.add_scalar('{}/{}_{}'.format(metric_name,dataset_name,metric_name), metric_value, cur_epoch * epoch_steps + cur_step)
            if cur_step in [0, epoch_steps - 1] or  cur_step % (self.cfg.log_interval_step * 10) == 0: # first iter, last iter and several middle iter.
                for (name, parameter) in self.model.named_parameters():
                    # if torch.isnan(parameter).int().sum() > 0 : continue
                    self.tensorboard_writer.add_histogram(name, parameter, cur_epoch * epoch_steps + cur_step)
                    if parameter.requires_grad:
                        self.tensorboard_writer.add_histogram("%s_Grad"%name, parameter.grad, cur_epoch * epoch_steps + cur_step)

        metric_str = ";\t".join("{} = {:.4f}".format(metric_name, metric_value) for (metric_name, metric_value) in metric_values.items())
        if dataset_name == 'Train':
            out_str = '[{}] {} epoch({}) step ({}/{}) {}: {}'.format(dt, "rank(%s)"% ext_dist.my_rank if ext_dist.my_rank >=0 else "",
                                                                    cur_epoch,cur_step,epoch_steps,dataset_name,metric_str)
        else:
            out_str = '[{}] {} epoch({}) {}: {}'.format(dt,"rank(%s)"%ext_dist.my_rank if ext_dist.my_rank >=0 else "",
                                                        cur_epoch,dataset_name, metric_str)
        print(out_str)
        self.logger.info(out_str)

    def _dist_wrapper(self):
        """
            wrapper model for distributed training
        """            
        if ext_dist.my_size > 1:
            self.logger.info("training with DistributedDataParallel")
            self.model = ext_dist.DDP(self.model)
            self.model.loss = MethodType(lambda obj,*args: obj.module.loss(*args) ,self.model) #  obj stands for model, both original model and transferrable model
            if self.is_transferrable: # dispatch method to model.module
                self.model.get_backbone = MethodType(lambda obj: obj.module.get_backbone() ,self.model) # obj stands for model
                self.model.get_training_metrics = MethodType(lambda obj,*args: obj.module.get_training_metrics(*args) ,self.model) #  obj stands for model

    def _is_early_stop(self, metrics_map):
        """
            check whether training achieved pre-defined metric threshold
        """
        if ext_dist.my_rank <= 0 and self.early_stopping is not None: # non distributed training, or rank 0  in distributed training
            self.early_stopping(metrics_map[self.best_metric_name], self.best_metric_value)
            return self.early_stopping.early_stop
        else:
            return False

    def train_one_epoch(self, cur_epoch, epoch_steps):
        ''' train one epoch

        :param cur_epoch: current epoch
        :param epoch_steps: how many steps of an epoch
        :return:
        '''
        self.model.train()  # set training flag
        context = self.profiler if self.profiler is not None else contextlib.nullcontext()
        with context:
            for (cur_step,(data, label)) in enumerate(self.train_dataloader):
                ##### prepare data
                '''
                Four cases of data
                Case 1 - basic: input
                Case 2 - distiller with logits: (input, (logits,seed))
                Case 3 - adapter: (input1, input2)
                Case 4 - distiller with logits and adapter: ((input1, (logits1,seed1)),(input2, (logits2,seed2))) - too complex, to be supported
                '''
                if isinstance(data, torch.Tensor):
                    # case 1
                    data = data.to(self.device)
                    label = label.to(self.device)
                elif isinstance(data, Iterable):
                    # case 2
                    data[0] = data[0].to(self.device)
                    label[0] = label[0].to(self.device)
                    # case 3
                    if len(data)>=2 and isinstance(data[1], torch.Tensor):
                        assert len(data) == len(label), "data len[%s] must equal label len[%s]"%(len(data),len(label))
                        data[1] = data[1].to(self.device)
                        label[1] = label[1].to(self.device)
                else:
                    raise RuntimeError("unknown data type:%s"%type(data))

                ##### train and update optimizer/scheduler
                self.optimizer.zero_grad()
                output = self.model(data)
                loss_value = self.model.loss(output, label)
                loss_value.backward()
                self.optimizer.step()
                if self.warmup_scheduler and cur_epoch < self.cfg.warmup_scheduler_epoch:
                    self.warmup_scheduler.step()

                ##### calculate metrics and show results
                if cur_step % self.cfg.log_interval_step == 0:
                    if self.is_transferrable: 
                        metric_values = self.model.get_training_metrics(output, label, loss_value, self.metric)
                    else:
                        metric_values = {"loss": loss_value}
                        for (metric_name, metric_fn) in sorted(self.metric.items()):
                            metric_value = metric_fn(output, label)
                            metric_value = metric_value[0] if isinstance(metric_value, list) else metric_value
                            metric_values[metric_name] = metric_value
                    self.show_update_tensorboard_metric('Train', metric_values, cur_epoch, cur_step, epoch_steps)
                
                ##### update profiler
                if context is self.profiler:
                    context.step()

    def evaluate(self, cur_epoch=0, epoch_steps=1, test_flag=True):
        ''' evaluate epoch
        :param cur_epoch : current epoch
        :param epoch_steps : steps of one epoch
        :param test_flag : whether is test or validation
        :return: metric_value_maps
        '''
        datasetName = 'Test' if test_flag else 'Validation'
        context = self.profiler if self.profiler is not None else contextlib.nullcontext()
        with torch.no_grad():
            with context:
                self.backbone.eval()  # set evaluating flag
                loss_value = 0
                metric_values = {}
                sample_num = 0
                #################### iterate on dataset ##############
                for (cur_step,(data, label)) in enumerate(self.eval_dataloader):
                    data = data.to(self.device)
                    label = label.to(self.device)
                    output = self.backbone(data)
                    output = output.logits if self.cfg.model_type.startswith("huggingface") else output
                    if isinstance(output, Iterable):
                        if not isinstance(output, torch.Tensor): # Tensor is Iterable
                            output = output[0]
                    else:
                        raise RuntimeError("Known data type:%s"%type(data))

                    batch_size = data.size(0)
                    sample_num += batch_size
                    loss_value += self.backbone.loss(output, label).item() * batch_size
                    for (metric_name,metric_fn) in self.metric.items():
                        metric_value = metric_fn(output,label)
                        metric_value = metric_value[0] if isinstance(metric_value, list) else metric_value
                        if metric_name not in metric_values:
                            metric_values[metric_name] = metric_value * batch_size
                        else:
                            metric_values[metric_name] += metric_value * batch_size
                    if cur_step % self.cfg.log_interval_step == 0:
                        dt = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S') # date time
                        print(f"{dt} {cur_step}/{epoch_steps}")
                        self.logger.info(f"{dt} {cur_step}/{epoch_steps}")
                    if context is self.profiler:
                        context.step()
                ############## average ###################
                metric_values['loss'] = loss_value
                for metric_name in sorted(metric_values.keys()):
                    metric_values[metric_name] /= sample_num
                ############## show and update tensorboard metric ###################
                self.show_update_tensorboard_metric(datasetName,metric_values,cur_epoch,cur_step=0,epoch_steps=epoch_steps)
            return metric_values

    def save_checkpoints(self, epoch=0, update_best=False, model_dir=""):
        state = {
            "epoch": epoch,
            "best_metric": self.best_metric_value,
            "model": self.model.state_dict(),
            "optimizer": self.optimizer.state_dict(),
            "scheduler": self.scheduler.state_dict()}
        torch.save(state, os.path.join(model_dir, "latest.pth"))
        torch.save(self.backbone.state_dict(), os.path.join(model_dir, "backbone_latest.pth"))
        if epoch % self.cfg.model_save_interval == 0:
            torch.save(state, os.path.join(model_dir, f"epoch_{epoch}.pth"))
            torch.save(self.backbone.state_dict(), os.path.join(model_dir, f"backbone_epoch_{epoch}.pth"))
        if update_best:
            torch.save(state, os.path.join(model_dir,"best.pth"))
            torch.save(self.backbone.state_dict(), os.path.join(model_dir,"backbone_best.pth"))

    def fit(self, epoch_steps, model_dir, resume=False):
        ''' train function, and save the best trained model to model_dir
        
        :param epoch_steps: steps per epoch
        :param model_dir: model saved dir
        :param resume: flag for whether resume pretrained model
        :return: validation metric. If has Earlystopping, using the best metric; Else, using the last metric
        '''
        ##### pre_process
        self._pre_process()
        very_start_time = time.time()
        initial_epoch = self.cfg.start_epoch
        train_dataset = self.train_dataloader.dataset

        ##### resume saved model
        if resume:
            latest_model_path = os.path.join(model_dir,"latest.pth")
            print(f"Resume checkpoint from {latest_model_path}")
            self.logger.info(f"Resume checkpoint from {latest_model_path}")

            state = torch.load(latest_model_path, map_location=self.device)
            initial_epoch = state["epoch"] + 1
            self.best_metric_value = state["best_metric"]
            self.model.load_state_dict(state["model"])
            self.optimizer.load_state_dict(state["optimizer"])
            self.scheduler.load_state_dict(state["scheduler"])
        
        ##### start training
        for epoch in range(initial_epoch, self.cfg.train_epochs):
            start_time = time.time()
            update_best = False
            
            ##### set epoch for dataset (used for distiller save logits function)
            if hasattr(train_dataset, 'set_epoch'):
                train_dataset.set_epoch(epoch)

            ##### get learning rate
            if self.scheduler is not None:
                if type(self.scheduler) is torch.optim.lr_scheduler.ReduceLROnPlateau:
                    last_lr = [item['lr'] for item in self.scheduler.optimizer.state_dict()['param_groups']]
                else:
                    last_lr = self.scheduler.get_last_lr()
                print("Epoch [%s] learning rate: %s" % (epoch, last_lr))
                self.logger.info("Epoch [%s] learning rate: %s" % (epoch, last_lr))

            ###### train and evaluate for on epoch
            self.train_one_epoch(epoch, epoch_steps)
            metrics_map = self.evaluate(epoch,epoch_steps,test_flag=False)
            
            ###### flush tensorboard and update scheduler
            if self.tensorboard_writer is not None:
                self.tensorboard_writer.flush()
            if not (self.warmup_scheduler and epoch < self.cfg.warmup_scheduler_epoch):
                if self.cfg.lr_scheduler == "ReduceLROnPlateau":
                    self.scheduler.step(metrics_map[self.best_metric_name])
                else:
                    self.scheduler.step()
                
            ###### update best metric value
            if metrics_map[self.best_metric_name] >= self.best_metric_value:
                self.best_metric_value = metrics_map[self.best_metric_name]
                update_best = True
                print(f"Best Epoch: {epoch}, {self.best_metric_name}: {self.best_metric_value}")
                self.logger.info(f"Best Epoch: {epoch}, {self.best_metric_name}: {self.best_metric_value}")
           
            ###### save checkpoint
            if ext_dist.my_rank <= 0: # non distributed training, or rank 0  in distributed training
                self.save_checkpoints(epoch, update_best, model_dir)

            print(f"Epoch {epoch} took {time.time()-start_time} seconds")
            self.logger.info(f"Epoch {epoch} took {time.time()-start_time} seconds")

            ###### check early stoppping
            if self._is_early_stop(metrics_map):
                self.logger.warning("Early stop after epoch:%s, the best %s is %s" % (epoch, self.best_metric_name, self.best_metric_value))
                break

        self.logger.info(F"Total time:{time.time() - very_start_time}")
        self._post_process()
        return self.best_metric_value 


#    Copyright 2022, Intel Corporation.
#    Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from torch import nn
import torch
import torch.nn.functional
from nnunet.network_architecture.initialization import InitWeights_He
from nnunet.network_architecture.generic_UNet import (ConvDropoutNormNonlin, StackedConvLayers)



class SegDiscriminator(nn.Module):
    BASE_NUM_FILTERS = 50

    def __init__(self, input_channels, threeD=True, 
                 pool_op_kernel_sizes=None,
                 ):

        super(SegDiscriminator, self).__init__()

        if threeD:
            conv_op = nn.Conv3d
            dropout_op = nn.Dropout3d
            norm_op = nn.InstanceNorm3d
        else:
            conv_op = nn.Conv2d
            dropout_op = nn.Dropout2d
            norm_op = nn.InstanceNorm2d
         
        nonlin=nn.LeakyReLU
        weightInitializer=InitWeights_He(1e-2)
        basic_block=ConvDropoutNormNonlin
        num_conv_per_stage=2

        nonlin_kwargs = {'negative_slope': 1e-2, 'inplace': True}
        # dropout_op_kwargs = {'p': 0.5, 'inplace': True}
        dropout_op_kwargs = {'p': 0, 'inplace': True}
        # norm_op_kwargs = {'eps': 1e-5, 'affine': True, 'momentum': 0.1}
        norm_op_kwargs = {'eps': 1e-5, 'affine': True}
        conv_kwargs = {'stride': 1, 'dilation': 1, 'bias': True}

        self.input_channels = input_channels
        self.conv_kwargs = conv_kwargs
        self.nonlin = nonlin
        self.nonlin_kwargs = nonlin_kwargs
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.weightInitializer = weightInitializer
        self.conv_op = conv_op
        self.norm_op = norm_op
        self.dropout_op = dropout_op

        if conv_op == nn.Conv2d:
            if pool_op_kernel_sizes is None:
                pool_op_kernel_sizes = [(2, 2)] * len(input_channels)
            self.avgpool = nn.AdaptiveAvgPool2d(1)
            conv_kernel_sizes = (3, 3)
            conv_pad_sizes = [1, 1]
        elif conv_op == nn.Conv3d:
            if pool_op_kernel_sizes is None:
                pool_op_kernel_sizes = [(2, 2, 2)] * len(input_channels)
            self.avgpool = nn.AdaptiveAvgPool3d(1)
            conv_kernel_sizes = (3, 3, 3)
            conv_pad_sizes = [1, 1, 1]
        else:
            raise ValueError("unknown convolution dimensionality, conv op: %s" % str(conv_op))

        self.conv_kwargs['kernel_size'] = conv_kernel_sizes
        self.conv_kwargs['padding'] = conv_pad_sizes

        self.conv_blocks_context = []
        for d in range(len(input_channels)):

            if d == 0:
                input_features = input_channels[d]
                output_features = self.BASE_NUM_FILTERS
            else:
                input_features = d * self.BASE_NUM_FILTERS + input_channels[d]
                output_features = (d+1) * self.BASE_NUM_FILTERS
            
            first_stride = pool_op_kernel_sizes[d]

            # add convolutions
            self.conv_blocks_context.append(StackedConvLayers(input_features, output_features, num_conv_per_stage,
                                                              self.conv_op, self.conv_kwargs, self.norm_op,
                                                              self.norm_op_kwargs, self.dropout_op,
                                                              self.dropout_op_kwargs, self.nonlin, self.nonlin_kwargs,
                                                              first_stride, basic_block=basic_block))

        # register all modules properly
        self.conv_blocks_context = nn.ModuleList(self.conv_blocks_context)
        if self.weightInitializer is not None:
            self.apply(self.weightInitializer)
        
    def forward(self, features):
        if not isinstance(features, list) and not isinstance(features, tuple):
            features = [features]

        input = features[0]
        x = self.conv_blocks_context[0](input)

        for i in range(1, len(self.input_channels)):
            x = torch.cat((x, features[i]), dim=1)
            x = self.conv_blocks_context[i](x)
   
        x = self.avgpool(x)
        x = x.view(x.size(0) * x.size(1), -1)

        return x
        

class FCDiscriminator(nn.Module):
    BASE_NUM_FILTERS = 50
    CONV_NUM = 4

    def __init__(self):
        super().__init__()

        input_channel=1
        num_conv_per_stage=2
        pool_op_kernel_size = (2, 2)
        first_stride = pool_op_kernel_size
        conv_kernel_sizes = (3, 3)
        conv_pad_sizes = [1, 1]
        weightInitializer=InitWeights_He(1e-2)
        basic_block=ConvDropoutNormNonlin
        avgpool = nn.AdaptiveAvgPool2d(1)

        conv_op = nn.Conv2d
        norm_op = nn.BatchNorm2d
        dropout_op = nn.Dropout2d
        nonlin=nn.LeakyReLU

        nonlin_kwargs = {'negative_slope': 1e-2, 'inplace': True}
        # dropout_op_kwargs = {'p': 0.5, 'inplace': True}
        dropout_op_kwargs = {'p': 0, 'inplace': True}
        # norm_op_kwargs = {'eps': 1e-5, 'affine': True, 'momentum': 0.1}
        norm_op_kwargs = {'eps': 1e-5, 'affine': True}
        conv_kwargs = {'stride': 1, 'dilation': 1, 'bias': True}

        self.conv_kwargs = conv_kwargs
        self.nonlin = nonlin
        self.nonlin_kwargs = nonlin_kwargs
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.weightInitializer = weightInitializer
        self.conv_op = conv_op
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.avgpool = avgpool
        self.conv_kwargs['kernel_size'] = conv_kernel_sizes
        self.conv_kwargs['padding'] = conv_pad_sizes

        self.conv_blocks_context = []
        for d in range(self.CONV_NUM):

            if d == 0:
                input_features = input_channel
            else:
                input_features = d * self.BASE_NUM_FILTERS
            output_features = (d+1) * self.BASE_NUM_FILTERS

            # add convolutions
            self.conv_blocks_context.append(StackedConvLayers(input_features, output_features, num_conv_per_stage,
                                                              self.conv_op, self.conv_kwargs, self.norm_op,
                                                              self.norm_op_kwargs, self.dropout_op,
                                                              self.dropout_op_kwargs, self.nonlin, self.nonlin_kwargs,
                                                              first_stride, basic_block=basic_block))

        # register all modules properly
        self.conv_blocks_context = nn.ModuleList(self.conv_blocks_context)
        if self.weightInitializer is not None:
            self.apply(self.weightInitializer)

        self.final_layer = nn.Sequential(
            nn.Linear(output_features, 1),
            nn.Sigmoid()
        )
        
    def forward(self, x):
        x = torch.flatten(x, start_dim=0, end_dim=2)
        x = torch.unsqueeze(x, 1)

        for i in range(self.CONV_NUM):
            x = self.conv_blocks_context[i](x)
   
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.final_layer(x)

        return x



import torch.nn.functional as F

from e2eAIOK.DeNas.module.embedding_base import EmbeddingBase


class Embedding(EmbeddingBase):
    """
    Computes an embedding x = wx.
    """

    def __init__(self, num_embeddings, embedding_dim, blank_id=0):

        super().__init__(num_embeddings, embedding_dim, blank_id=0)

    def forward(self, x):
        return F.embedding(
            x, self.weight, self.padding_idx, self.max_norm,
            self.norm_type, self.scale_grad_by_freq, self.sparse)



from torch.utils.data import Dataset
from e2eAIOK.common.trainer.data.data_utils.loader import rgb_loader,l_loader
import logging
import os

class Office31(Dataset):
    ''' Office31 Dataset

    '''
    def __init__(self, data_path, label_map, transform,img_mode='RGB'):
        ''' Init method

        :param data_path: img data location
        :param label_map: map from label name to label id
        :param transform: data transform
        :param img_mode: img mode
        '''
        self.data_path = data_path
        self._label_map = label_map
        self.transform = transform
        self.img_mode = img_mode
        self._getFileAndImgPath(data_path)

        if img_mode.upper() == 'RGB':
            self.loader = rgb_loader
        elif img_mode.upper()  == 'L':
            self.loader = l_loader
        else:
            logging.error("mode muse be one of: 'RGB','L' (ignore letter case), but found :%s"%img_mode)
            raise ValueError("mode muse be one of: 'RGB','L' (ignore letter case), but found :%s"%img_mode)

    def _getFileAndImgPath(self,root_path):
        self.imgs = []

        for label_name in sorted(os.listdir(root_path)):
            label_path = "%s/%s"%(root_path,label_name)
            label_id = self._label_map[label_name]
            for img_name in os.listdir(label_path):
                img_path = "%s/%s"%(label_path,img_name)
                self.imgs.append((img_path,label_id))

    def __getitem__(self, index):
        path, label = self.imgs[index]
        img = self.loader(path)
        if self.transform is not None:
            img = self.transform(img)

        return img, label

    def __len__(self):
        return len(self.imgs)

    def __str__(self):
        return 'Office31: image num [%s], img_mode[%s], transform [%s]'%(
            len(self.imgs),self.img_mode,self.transform)



import torch
from e2eAIOK.DeNas.asr.data.dataio.dataio import length_to_mask


def Accuracy(log_probabilities, targets, length=None):
    """Calculates the accuracy for predicted log probabilities and targets in a batch.

    Arguments
    ----------
    log_probabilities : tensor
        Predicted log probabilities (batch_size, time, feature).
    targets : tensor
        Target (batch_size, time).
    length : tensor
        Length of target (batch_size,).
    """
    if length is not None:
        mask = length_to_mask(
            length * targets.shape[1], max_len=targets.shape[1],
        ).bool()
        if len(targets.shape) == 3:
            mask = mask.unsqueeze(2).repeat(1, 1, targets.shape[2])

    padded_pred = log_probabilities.argmax(-1)

    if length is not None:
        numerator = torch.sum(
            padded_pred.masked_select(mask) == targets.masked_select(mask)
        )
        denominator = torch.sum(mask)
    else:
        numerator = torch.sum(padded_pred == targets)
        denominator = targets.shape[1]
    return float(numerator), float(denominator)


class AccuracyStats:
    """Module for calculate the overall one-step-forward prediction accuracy.
    """

    def __init__(self):
        self.correct = 0
        self.total = 0

    def append(self, log_probabilities, targets, length=None):
        """This function is for updating the stats according to the prediction
        and target in the current batch.

        Arguments
        ----------
        log_probabilities : tensor
            Predicted log probabilities (batch_size, time, feature).
        targets : tensor
            Target (batch_size, time).
        length: tensor
            Length of target (batch_size,).
        """
        numerator, denominator = Accuracy(log_probabilities, targets, length)
        self.correct += numerator
        self.total += denominator

    def summarize(self):
        """Computes the accuract metric."""
        return self.correct / self.total



import collections

EDIT_SYMBOLS = {
    "eq": "=",  # when tokens are equal
    "ins": "I",
    "del": "D",
    "sub": "S",
}


def op_table(a, b):
    """Table of edit operations between a and b.

    Solves for the table of edit operations, which is mainly used to
    compute word error rate. The table is of size ``[|a|+1, |b|+1]``,
    and each point ``(i, j)`` in the table has an edit operation. The
    edit operations can be deterministically followed backwards to
    find the shortest edit path to from ``a[:i-1] to b[:j-1]``. Indexes
    of zero (``i=0`` or ``j=0``) correspond to an empty sequence.

    The algorithm itself is well known, see

    `Levenshtein distance <https://en.wikipedia.org/wiki/Levenshtein_distance>`_

    Note that in some cases there are multiple valid edit operation
    paths which lead to the same edit distance minimum.

    Arguments
    ---------
    a : iterable
        Sequence for which the edit operations are solved.
    b : iterable
        Sequence for which the edit operations are solved.

    Returns
    -------
    list
        List of lists, Matrix, Table of edit operations.
    """
    # For the dynamic programming algorithm, only two rows are really needed:
    # the one currently being filled in, and the previous one
    # The following is also the right initialization
    prev_row = [j for j in range(len(b) + 1)]
    curr_row = [0] * (len(b) + 1)  # Just init to zero
    # For the edit operation table we will need the whole matrix.
    # We will initialize the table with no-ops, so that we only need to change
    # where an edit is made.
    table = [
        [EDIT_SYMBOLS["eq"] for j in range(len(b) + 1)]
        for i in range(len(a) + 1)
    ]
    # We already know the operations on the first row and column:
    for i in range(len(a) + 1):
        table[i][0] = EDIT_SYMBOLS["del"]
    for j in range(len(b) + 1):
        table[0][j] = EDIT_SYMBOLS["ins"]
    table[0][0] = EDIT_SYMBOLS["eq"]
    # The rest of the table is filled in row-wise:
    for i, a_token in enumerate(a, start=1):
        curr_row[0] += 1  # This trick just deals with the first column.
        for j, b_token in enumerate(b, start=1):
            # The dynamic programming algorithm cost rules
            insertion_cost = curr_row[j - 1] + 1
            deletion_cost = prev_row[j] + 1
            substitution = 0 if a_token == b_token else 1
            substitution_cost = prev_row[j - 1] + substitution
            # Here copying the Kaldi compute-wer comparison order, which in
            # ties prefers:
            # insertion > deletion > substitution
            if (
                substitution_cost < insertion_cost
                and substitution_cost < deletion_cost
            ):
                curr_row[j] = substitution_cost
                # Again, note that if not substitution, the edit table already
                # has the correct no-op symbol.
                if substitution:
                    table[i][j] = EDIT_SYMBOLS["sub"]
            elif deletion_cost < insertion_cost:
                curr_row[j] = deletion_cost
                table[i][j] = EDIT_SYMBOLS["del"]
            else:
                curr_row[j] = insertion_cost
                table[i][j] = EDIT_SYMBOLS["ins"]
        # Move to the next row:
        prev_row[:] = curr_row[:]
    return table


def alignment(table):
    """Get the edit distance alignment from an edit op table.

    Walks back an edit operations table, produced by calling ``table(a, b)``,
    and collects the edit distance alignment of a to b. The alignment
    shows which token in a corresponds to which token in b. Note that the
    alignment is monotonic, one-to-zero-or-one.

    Arguments
    ----------
    table : list
        Edit operations table from ``op_table(a, b)``.

    Returns
    -------
    list
        Schema: ``[(str <edit-op>, int-or-None <i>, int-or-None <j>),]``
        List of edit operations, and the corresponding indices to a and b.
        See the EDIT_SYMBOLS dict for the edit-ops.
        The i indexes a, j indexes b, and the indices can be None, which means
        aligning to nothing.
    """
    # The alignment will be the size of the longer sequence.
    # form: [(op, a_index, b_index)], index is None when aligned to empty
    alignment = []
    # Now we'll walk back the table to get the alignment.
    i = len(table) - 1
    j = len(table[0]) - 1
    while not (i == 0 and j == 0):
        if i == 0:
            j -= 1
            alignment.insert(0, (EDIT_SYMBOLS["ins"], None, j))
        elif j == 0:
            i -= 1
            alignment.insert(0, (EDIT_SYMBOLS["del"], i, None))
        else:
            if table[i][j] == EDIT_SYMBOLS["ins"]:
                j -= 1
                alignment.insert(0, (EDIT_SYMBOLS["ins"], None, j))
            elif table[i][j] == EDIT_SYMBOLS["del"]:
                i -= 1
                alignment.insert(0, (EDIT_SYMBOLS["del"], i, None))
            elif table[i][j] == EDIT_SYMBOLS["sub"]:
                i -= 1
                j -= 1
                alignment.insert(0, (EDIT_SYMBOLS["sub"], i, j))
            else:
                i -= 1
                j -= 1
                alignment.insert(0, (EDIT_SYMBOLS["eq"], i, j))
    return alignment


def count_ops(table):
    """Count the edit operations in the shortest edit path in edit op table.

    Walks back an edit operations table produced by table(a, b) and
    counts the number of insertions, deletions, and substitutions in the
    shortest edit path. This information is typically used in speech
    recognition to report the number of different error types separately.

    Arguments
    ----------
    table : list
        Edit operations table from ``op_table(a, b)``.

    Returns
    -------
    collections.Counter
        The counts of the edit operations, with keys:

        * "insertions"
        * "deletions"
        * "substitutions"

        NOTE: not all of the keys might appear explicitly in the output,
        but for the missing keys collections. The counter will return 0.
    """
    edits = collections.Counter()
    # Walk back the table, gather the ops.
    i = len(table) - 1
    j = len(table[0]) - 1
    while not (i == 0 and j == 0):
        if i == 0:
            edits["insertions"] += 1
            j -= 1
        elif j == 0:
            edits["deletions"] += 1
            i -= 1
        else:
            if table[i][j] == EDIT_SYMBOLS["ins"]:
                edits["insertions"] += 1
                j -= 1
            elif table[i][j] == EDIT_SYMBOLS["del"]:
                edits["deletions"] += 1
                i -= 1
            else:
                if table[i][j] == EDIT_SYMBOLS["sub"]:
                    edits["substitutions"] += 1
                i -= 1
                j -= 1
    return edits


def _batch_to_dict_format(ids, seqs):
    # Used by wer_details_for_batch
    return dict(zip(ids, seqs))


def wer_details_for_batch(ids, refs, hyps, compute_alignments=False):
    """Convenient batch interface for ``wer_details_by_utterance``.

    ``wer_details_by_utterance`` can handle missing hypotheses, but
    sometimes (e.g. CTC training with greedy decoding) they are not needed,
    and this is a convenient interface in that case.

    Arguments
    ---------
    ids : list, torch.tensor
        Utterance ids for the batch.
    refs : list, torch.tensor
        Reference sequences.
    hyps : list, torch.tensor
        Hypothesis sequences.
    compute_alignments : bool, optional
        Whether to compute alignments or not. If computed, the details
        will also store the refs and hyps. (default: False)

    Returns
    -------
    list
        See ``wer_details_by_utterance``
    """
    refs = _batch_to_dict_format(ids, refs)
    hyps = _batch_to_dict_format(ids, hyps)
    return wer_details_by_utterance(
        refs, hyps, compute_alignments=compute_alignments, scoring_mode="strict"
    )


def wer_details_by_utterance(
    ref_dict, hyp_dict, compute_alignments=False, scoring_mode="strict"
):
    """Computes a wealth WER info about each single utterance.

    This info can then be used to compute summary details (WER, SER).

    Arguments
    ---------
    ref_dict : dict
        Should be indexable by utterance ids, and return the reference tokens
        for each utterance id as iterable
    hyp_dict : dict
        Should be indexable by utterance ids, and return
        the hypothesis tokens for each utterance id as iterable
    compute_alignments : bool
        Whether alignments should also be saved.
        This also saves the tokens themselves, as they are probably
        required for printing the alignments.
    scoring_mode : {'strict', 'all', 'present'}
        How to deal with missing hypotheses (reference utterance id
        not found in hyp_dict).

        * 'strict': Raise error for missing hypotheses.
        * 'all': Score missing hypotheses as empty.
        * 'present': Only score existing hypotheses.

    Returns
    -------
    list
        A list with one entry for every reference utterance. Each entry is a
        dict with keys:

        * "key": utterance id
        * "scored": (bool) Whether utterance was scored.
        * "hyp_absent": (bool) True if a hypothesis was NOT found.
        * "hyp_empty": (bool) True if hypothesis was considered empty
          (either because it was empty, or not found and mode 'all').
        * "num_edits": (int) Number of edits in total.
        * "num_ref_tokens": (int) Number of tokens in the reference.
        * "WER": (float) Word error rate of the utterance.
        * "insertions": (int) Number of insertions.
        * "deletions": (int) Number of deletions.
        * "substitutions": (int) Number of substitutions.
        * "alignment": If compute_alignments is True, alignment as list
          If compute_alignments is False, this is None.
        * "ref_tokens": (iterable) The reference tokens
          only saved if alignments were computed, else None.
        * "hyp_tokens": (iterable) the hypothesis tokens,
          only saved if alignments were computed, else None.

    Raises
    ------
    KeyError
        If scoring mode is 'strict' and a hypothesis is not found.
    """
    details_by_utterance = []
    for key, ref_tokens in ref_dict.items():
        # Initialize utterance_details
        utterance_details = {
            "key": key,
            "scored": False,
            "hyp_absent": None,
            "hyp_empty": None,
            "num_edits": None,
            "num_ref_tokens": len(ref_tokens),
            "WER": None,
            "insertions": None,
            "deletions": None,
            "substitutions": None,
            "alignment": None,
            "ref_tokens": ref_tokens if compute_alignments else None,
            "hyp_tokens": None,
        }
        if key in hyp_dict:
            utterance_details.update({"hyp_absent": False})
            hyp_tokens = hyp_dict[key]
        elif scoring_mode == "all":
            utterance_details.update({"hyp_absent": True})
            hyp_tokens = []
        elif scoring_mode == "present":
            utterance_details.update({"hyp_absent": True})
            details_by_utterance.append(utterance_details)
            continue  # Skip scoring this utterance
        elif scoring_mode == "strict":
            raise KeyError(
                "Key "
                + key
                + " in reference but missing in hypothesis and strict mode on."
            )
        else:
            raise ValueError("Invalid scoring mode: " + scoring_mode)
        # Compute edits for this utterance
        table = op_table(ref_tokens, hyp_tokens)
        ops = count_ops(table)
        # Update the utterance-level details if we got this far:
        utterance_details.update(
            {
                "scored": True,
                "hyp_empty": True
                if len(hyp_tokens) == 0
                else False,  # This also works for e.g. torch tensors
                "num_edits": sum(ops.values()),
                "num_ref_tokens": len(ref_tokens),
                "WER": 100.0 * sum(ops.values()) / len(ref_tokens),
                "insertions": ops["insertions"],
                "deletions": ops["deletions"],
                "substitutions": ops["substitutions"],
                "alignment": alignment(table) if compute_alignments else None,
                "ref_tokens": ref_tokens if compute_alignments else None,
                "hyp_tokens": hyp_tokens if compute_alignments else None,
            }
        )
        details_by_utterance.append(utterance_details)
    return details_by_utterance


def wer_summary(details_by_utterance):
    """
    Computes summary stats from the output of details_by_utterance

    Summary stats like WER

    Arguments
    ---------
    details_by_utterance : list
        See the output of wer_details_by_utterance

    Returns
    -------
    dict
        Dictionary with keys:

        * "WER": (float) Word Error Rate.
        * "SER": (float) Sentence Error Rate (percentage of utterances
          which had at least one error).
        * "num_edits": (int) Total number of edits.
        * "num_scored_tokens": (int) Total number of tokens in scored
          reference utterances (a missing hypothesis might still
          have been scored with 'all' scoring mode).
        * "num_erraneous_sents": (int) Total number of utterances
          which had at least one error.
        * "num_scored_sents": (int) Total number of utterances
          which were scored.
        * "num_absent_sents": (int) Hypotheses which were not found.
        * "num_ref_sents": (int) Number of all reference utterances.
        * "insertions": (int) Total number of insertions.
        * "deletions": (int) Total number of deletions.
        * "substitutions": (int) Total number of substitutions.

        NOTE: Some cases lead to ambiguity over number of
        insertions, deletions and substitutions. We
        aim to replicate Kaldi compute_wer numbers.
    """
    # Build the summary details:
    ins = dels = subs = 0
    num_scored_tokens = (
        num_scored_sents
    ) = num_edits = num_erraneous_sents = num_absent_sents = num_ref_sents = 0
    for dets in details_by_utterance:
        num_ref_sents += 1
        if dets["scored"]:
            num_scored_sents += 1
            num_scored_tokens += dets["num_ref_tokens"]
            ins += dets["insertions"]
            dels += dets["deletions"]
            subs += dets["substitutions"]
            num_edits += dets["num_edits"]
            if dets["num_edits"] > 0:
                num_erraneous_sents += 1
        if dets["hyp_absent"]:
            num_absent_sents += 1
    wer_details = {
        "WER": 100.0 * num_edits / num_scored_tokens,
        "SER": 100.0 * num_erraneous_sents / num_scored_sents,
        "num_edits": num_edits,
        "num_scored_tokens": num_scored_tokens,
        "num_erraneous_sents": num_erraneous_sents,
        "num_scored_sents": num_scored_sents,
        "num_absent_sents": num_absent_sents,
        "num_ref_sents": num_ref_sents,
        "insertions": ins,
        "deletions": dels,
        "substitutions": subs,
    }
    return wer_details



import subprocess
import yaml
import logging
import time


from e2eAIOK.common.utils import *
from e2eAIOK.utils.hydroconfig import default_settings
from e2eAIOK.SDA.modeladvisor.BaseModelAdvisor import BaseModelAdvisor

class TestAdvisor(BaseModelAdvisor):
    def __init__(self, dataset_meta_path, train_path, eval_path, settings):
        super().__init__(dataset_meta_path, train_path, eval_path, settings)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')

    # ====== Implementation of required methods ======

    def update_metrics(self):
        metrics = []
        metrics.append({'name': 'accuracy', 'value': self.mean_accuracy})
        metrics.append({'name': 'training_time', 'value': self.training_time})
        self.params['model_parameter']['metrics'] = metrics
        return self.params['model_parameter']['metrics']

    def initialize_model_parameter(self, assignments=None):
        config = {}
        tuned_parameters = {}
        if assignments:
            tuned_parameters['max_depth'] = assignments['max_depth']
            tuned_parameters['learning_rate'] = assignments['learning_rate']
            tuned_parameters['min_split_loss'] = assignments['min_split_loss']
        else:
            tuned_parameters['max_depth'] = 11
            tuned_parameters['learning_rate'] = float(0.9294458527831317)
            tuned_parameters['min_split_loss'] = float(6.88375281543753)
        config['tuned_parameters'] = tuned_parameters
        self.params['model_parameter'] = config
        self.params['model_saved_path'] = os.path.join(
            self.params['save_path'], get_hash_string(str(tuned_parameters)))
        return config

    def generate_sigopt_yaml(self, file='test_sigopt.yaml'):
        config = {}
        config['project'] = 'e2eaiok'
        config['experiment'] = 'sklearn'
        parameters = [{
            'name': 'max_depth',
            'bounds': {
                'min': 3,
                'max': 12
            },
            'type': 'int'
        }, {
            'name': 'learning_rate',
            'bounds': {
                'min': 0.0,
                'max': 1.0
            },
            'type': 'double'
        }, {
            'name': 'min_split_loss',
            'bounds': {
                'min': 0.0,
                'max': 10
            },
            'type': 'double'
        }]
        user_defined_parameter = self.params['model_parameter'][
            'parameters'] if ('model_parameter' in self.params) and (
                'parameters' in self.params['model_parameter']) else None
        config['parameters'] = parameters
        if user_defined_parameter:
            self.logger.info(
                f"Update with user defined parameters {user_defined_parameter}"
            )
            update_list(config['parameters'], user_defined_parameter)
        config['metrics'] = [{
            'name': 'accuracy',
            'strategy': 'optimize',
            'objective': 'maximize'
        }, {
            'name': 'training_time',
            'objective': 'minimize'
        }]
        user_defined_metrics = self.params['model_parameter']['metrics'] if (
            'model_parameter' in self.params) and (
                'metrics' in self.params['model_parameter']) else None
        if user_defined_metrics:
            self.logger.info(
                f"Update with user defined parameters {user_defined_metrics}")
            update_list(config['metrics'], user_defined_metrics)
        config['observation_budget'] = self.params['observation_budget']

        # TODO: Add all parameter tuning here

        saved_path = os.path.join(self.params['save_path'], file)
        with open(saved_path, 'w') as f:
            yaml.dump(config, f)
        return config

    def train_model(self, args):
        start_time = time.time()
        self.mean_accuracy, model_path = self.dist_launch(args)
        self.training_time = time.time() - start_time
        metrics = self.update_metrics()
        return self.training_time, model_path, metrics

    def dist_launch(self, args):
        # construct WnD launch command with mpi
        max_depth = args['model_parameter']["tuned_parameters"]['max_depth']
        learning_rate = args['model_parameter']["tuned_parameters"][
            'learning_rate']
        min_split_loss = args['model_parameter']["tuned_parameters"][
            'min_split_loss']
        model_saved_path = args['model_saved_path']
        cmd = []
        cmd.append(f"python")
        cmd.append(f"example/sklearn_train.py")
        cmd.append(f"--max_depth")
        cmd.append(f"{max_depth}")
        cmd.append(f"--learning_rate")
        cmd.append(f"{learning_rate}")
        cmd.append(f"--min_split_loss")
        cmd.append(f"{min_split_loss}")
        cmd.append(f"--saved_path")
        cmd.append(f"{model_saved_path}")
        self.logger.info(f'### Starting model training ###, launch cmd is: {" ".join(cmd)}')
        output = subprocess.check_output(cmd)
        return float(output), model_saved_path



'''
Copyright (C) 2010-2021 Alibaba Group Holding Limited.
'''

import os
import sys
import time
import torch

def get_model_latency(model, batch_size, resolution, in_channels, gpu, repeat_times, fp16):
    if gpu is not None:
        device = torch.device('cuda:{}'.format(gpu))
    else:
        device = torch.device('cpu')
    if fp16:
        model = model.half()
        dtype = torch.float16
    else:
        dtype = torch.float32
    the_image = torch.randn(batch_size, in_channels, resolution, resolution, dtype=dtype,
                            device=device)
    model.eval()
    warmup_T = 1
    with torch.no_grad():
        for i in range(warmup_T):
            the_output = model(the_image)
        start_timer = time.time()
        for repeat_count in range(repeat_times):
            the_output = model(the_image)
    end_timer = time.time()
    the_latency = (end_timer - start_timer) / float(repeat_times) / batch_size
    return the_latency


#!/usr/bin/python
# -*- coding: utf-8 -*-

from torch.utils.data import Dataset
from .loader import rgb_loader,l_loader
import logging

class ImageList(Dataset):
    ''' ImageList dataset

    '''
    def __init__(self, data_path,label_records,transform,img_mode='RGB'):
        ''' init method

        :param data_path: image dir
        :param label_records: label record list, each record format: "image_name image_label"
        :param transform: data transform
        :param img_mode: img mode, must one "RGB" or "L" (ignore letter case)
        '''
        imgs = [("%s/%s"%(data_path,val.split()[0]), int(val.split()[1])) for val in label_records]
        if len(imgs) == 0:
            logging.error("Found 0 images in subfolders of: %s" %data_path)
            raise (RuntimeError("Found 0 images in subfolders of: %s" %data_path))

        self.imgs = imgs
        self.transform = transform
        self.img_mode = img_mode

        if img_mode.upper() == 'RGB':
            self.loader = rgb_loader
        elif img_mode.upper()  == 'L':
            self.loader = l_loader
        else:
            logging.error("mode muse be one of: 'RGB','L' (ignore letter case), but found :%s"%img_mode)
            raise ValueError("mode muse be one of: 'RGB','L' (ignore letter case), but found :%s"%img_mode)

    def __getitem__(self, index):
        path, label = self.imgs[index]
        img = self.loader(path)
        if self.transform is not None:
            img = self.transform(img)
        return img, label

    def __len__(self):
        return len(self.imgs)

    def __str__(self):
        return 'ImageList: image num [%s], img_mode[%s], transform [%s]'%(
            len(self.imgs),self.img_mode,self.transform)


import os
import ast
import time
import random
import numpy as np
import json
import shutil
import torch
import torch.nn as nn
import torch.nn.functional as F

import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist
from e2eAIOK.DeNas.module.nlp.optimization import BertAdam
from e2eAIOK.DeNas.nlp.supernet_bert import CrossEntropyQALoss
from e2eAIOK.DeNas.nlp.utils_eval import do_qa_eval

from e2eAIOK.DeNas.module.nlp.layernorm_super import LayerNormSuper
from e2eAIOK.DeNas.module.nlp.Linear_super import LinearSuper
from thop.vision.basic_hooks import count_normalization, count_linear
from ptflops.pytorch_ops import linear_flops_counter_hook

def customer_ops_map_thop():
    customer_ops_map = {LayerNormSuper: count_normalization,
                        LinearSuper: count_linear}
    return customer_ops_map

def generate_search_space(search_space_config):
        # build arch space
        search_space = {}
        search_space['layer_num'] = range(int(search_space_config['LAYER_NUM']['bounds']['min']), int(search_space_config['LAYER_NUM']['bounds']['max'])+1)
        search_space['head_num'] = range(int(search_space_config['HEAD_NUM']['bounds']['min']), int(search_space_config['HEAD_NUM']['bounds']['max']) + int(search_space_config['HEAD_NUM']['bounds']['step']), int(search_space_config['HEAD_NUM']['bounds']['step']))
        search_space['hidden_size'] = range(int(search_space_config['HIDDEN_SIZE']['bounds']['min']), int(search_space_config['HIDDEN_SIZE']['bounds']['max']) + int(search_space_config['HIDDEN_SIZE']['bounds']['step']), int(search_space_config['HIDDEN_SIZE']['bounds']['step']))
        search_space['ffn_size'] = range(int(search_space_config['INTERMEDIATE_SIZE']['bounds']['min']), int(search_space_config['INTERMEDIATE_SIZE']['bounds']['max']) + int(search_space_config['INTERMEDIATE_SIZE']['bounds']['step']), int(search_space_config['INTERMEDIATE_SIZE']['bounds']['step']))
        return search_space

def get_subconfig(cand):
    subconfig = dict()
    subconfig['sample_layer_num'] = cand[0]
    subconfig['sample_num_attention_heads'] = [cand[1]] * cand[0]
    subconfig['sample_qkv_sizes'] = [cand[2]] * cand[0]
    subconfig['sample_hidden_size'] = cand[3]
    subconfig['sample_intermediate_sizes'] = [cand[4]] * cand[0]
    return subconfig

def decode_arch(arches_file):
    subbert_config = None
    with open(arches_file, 'r') as fin:
        for line in fin:
            line = line.strip()
            subbert_config = get_subconfig(ast.literal_eval(line))
    return subbert_config

def bert_populate_random_func(search_space):
    cand_tuple = list() #[layer_num, [num_attention_heads]*layer_num, [qkv_sizes]*layer_num, hidden_size, [intermediate_sizes]*layer_num]
    dimensions = ['head_num', 'hidden_size', 'ffn_size']
    depth = random.choice(search_space['layer_num'])
    cand_tuple.append(depth)
    for dimension in dimensions:
        if dimension == 'head_num':
            head_num = random.choice(search_space[dimension])
            qkv_size = head_num * 64
            cand_tuple.append(head_num)
            cand_tuple.append(qkv_size)
        elif dimension == 'hidden_size':
            cand_tuple.append(random.choice(search_space['hidden_size']))
        elif dimension == 'ffn_size':
            cand_tuple.append(random.choice(search_space['ffn_size']))

    return tuple(cand_tuple)

def bert_is_legal(cand, vis_dict, params, super_net):
    if cand not in vis_dict:
        vis_dict[cand] = {}
    info = vis_dict[cand]
    if 'visited' in info:
        return False
    subconfig = get_subconfig(cand)
    super_net.set_sample_config(subconfig)
    n_parameters = super_net.calc_sampled_param_num()
    info['params'] = n_parameters / 10.**6
    if info['params'] > params.max_param_limits:
        return False
    if info['params'] < params.min_param_limits:
        return False
    info['visited'] = True
    return True

def bert_mutation_random_func(m_prob, s_prob, search_space, top_candidates):
    cand = list(random.choice(top_candidates))
    depth, num_heads, qkv_sizes, hidden_size, ffn_sizes = cand[0], cand[1], cand[2], cand[3], cand[4]
    random_s = random.random()
    # depth
    if random_s < s_prob:
        new_depth = random.choice(search_space['layer_num'])
        depth = new_depth
        num_heads = random.choice(search_space['head_num'])
        qkv_sizes = num_heads * 64
        hidden_size = random.choice(search_space['hidden_size'])
        ffn_sizes = random.choice(search_space['ffn_size'])
    random_s = random.random()
    if random_s < m_prob:
        # num_heads
        num_heads = random.choice(search_space['head_num'])
        # qkv_sizes
        qkv_sizes = num_heads * 64
    # hidden_size
    random_s = random.random()
    if random_s < s_prob:
        hidden_size = random.choice(search_space['hidden_size'])
    # ffn_sizes
    random_s = random.random()
    if random_s < s_prob:
        ffn_sizes = random.choice(search_space['ffn_size'])

    result_cand = [depth] + [num_heads] + [qkv_sizes] + [hidden_size] + [ffn_sizes]
    return tuple(result_cand)

def bert_crossover_random_func(top_candidates):
    p1 = random.choice(top_candidates)
    p2 = random.choice(top_candidates)
    max_iters_tmp = 50
    while len(p1) != len(p2) and max_iters_tmp > 0:
        max_iters_tmp -= 1
        p1 = random.choice(top_candidates)
        p2 = random.choice(top_candidates)
    cand = []
    for ind, it in enumerate(zip(p1, p2)):
        if ind == 2:
            continue
        elif ind == 1:
            cand.append(random.choice(it))
            cand.append(cand[-1] * 64)
        else:
            cand.append(random.choice(it))
    return tuple(cand)

def get_bert_latency(model, batch_size, max_seq_length, gpu, infer_cnt):
    if gpu is None:
        device = 'cpu'
    else:
        device = 'cuda'
    input_ids = [9333] * max_seq_length
    input_masks = max_seq_length * [1]
    input_segments = max_seq_length * [0]
    input_ids = torch.tensor([input_ids]*batch_size, dtype=torch.long).to(device)
    input_masks = torch.tensor([input_masks]*batch_size, dtype=torch.long).to(device)
    input_segments = torch.tensor([input_segments]*batch_size, dtype=torch.long).to(device)

    aver_time = 0.
    model.eval()

    for i in range(int(infer_cnt)):
        start = time.time()
        with torch.no_grad():
            model.forward(input_ids, input_masks, input_segments)

        end = time.time()
        sep = 1000 * (end - start)

        if i == 0:
            continue
        else:
            aver_time += sep / (infer_cnt - 1)
    return aver_time

def bert_create_optimizer(model, cfg):
    if cfg.optimizer == "BertAdam":
        param_optimizer = list(model.named_parameters())
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        optimizer_grouped_parameters = [
            {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
            'weight_decay': cfg.weight_decay},
            {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
        ]
        num_train_optimization_steps = int(
            cfg.num_train_steps / cfg.gradient_accumulation_steps) * cfg.train_epochs
        if ext_dist.my_size > 1:
            num_train_optimization_steps = num_train_optimization_steps * ext_dist.my_size
        optimizer = BertAdam(optimizer_grouped_parameters,
                             schedule=cfg.lr_scheduler,
                             lr=cfg.learning_rate,
                             warmup=cfg.warmup_proportion,
                             t_total=num_train_optimization_steps,
                             weight_decay=cfg.weight_decay)
        return optimizer

def bert_create_criterion(cfg):
    if cfg.criterion == "CrossEntropyQALoss":
        criterion = CrossEntropyQALoss(cfg.max_seq_length)
    return criterion

def bert_create_scheduler(cfg):
    return None

def bert_create_metric(cfg):
    if cfg.eval_metric == "qa_f1":
        return do_qa_eval


#    Copyright 2022, Intel Corporation.
#    Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from torch import nn
import torch
import numpy as np
import torch.nn.functional
from nnunet.network_architecture.initialization import InitWeights_He
from nnunet.network_architecture.generic_UNet import ConvDropoutNormNonlin
from nnunet.training.loss_functions.deep_supervision import MultipleOutputLoss2
from nnunet.training.loss_functions.dice_loss import DC_and_CE_loss
from .generic_UNet import Generic_UNet


class Generic_UNet_DA(Generic_UNet):

    def __init__(self, threeD, input_channels, base_num_features, num_classes, 
                 num_conv_per_stage=2,  
                 pool_op_kernel_sizes=None,
                 conv_kernel_sizes=None,
                ):
        
        if threeD:
            conv_op = nn.Conv3d
            dropout_op = nn.Dropout3d
            norm_op = nn.InstanceNorm3d

        else:
            conv_op = nn.Conv2d
            dropout_op = nn.Dropout2d
            norm_op = nn.InstanceNorm2d
        
        self.threeD = threeD
        num_pool = len(pool_op_kernel_sizes)
        feat_map_mul_on_downscale=2
        nonlin = nn.LeakyReLU
        norm_op_kwargs = {'eps': 1e-5, 'affine': True}
        dropout_op_kwargs = {'p': 0, 'inplace': True}
        nonlin_kwargs = {'negative_slope': 1e-2, 'inplace': True}
        deep_supervision=True
        dropout_in_localization=False
        upscale_logits=False
        convolutional_pooling=True
        convolutional_upsampling=True
        weightInitializer=InitWeights_He(1e-2)
        final_nonlin = lambda x: x
        max_num_features=None
        basic_block=ConvDropoutNormNonlin
        seg_output_use_bias=False

        super().__init__(input_channels, base_num_features, num_classes, 
                num_pool, num_conv_per_stage=num_conv_per_stage,
                feat_map_mul_on_downscale=feat_map_mul_on_downscale, conv_op=conv_op,
                norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                dropout_op=dropout_op, dropout_op_kwargs=dropout_op_kwargs,
                nonlin=nonlin, nonlin_kwargs=nonlin_kwargs, deep_supervision=deep_supervision, 
                dropout_in_localization=dropout_in_localization,
                final_nonlin=final_nonlin, weightInitializer=weightInitializer, 
                pool_op_kernel_sizes=pool_op_kernel_sizes,
                conv_kernel_sizes=conv_kernel_sizes,
                upscale_logits=upscale_logits, convolutional_pooling=convolutional_pooling, 
                convolutional_upsampling=convolutional_upsampling,
                max_num_features=max_num_features, basic_block=basic_block,
                seg_output_use_bias=seg_output_use_bias)
        self.set_loss()

    def set_loss(self, batch_dice=True):
        self.loss = DC_and_CE_loss({'batch_dice': batch_dice, 'smooth': 1e-5, 'do_bg': False}, {})

        ################# Here we wrap the loss for deep supervision ############
        # we need to know the number of outputs of the network
        net_numpool = len(self.pool_op_kernel_sizes)

        # we give each output a weight which decreases exponentially (division by 2) as the resolution decreases
        # this gives higher resolution outputs more weight in the loss
        weights = np.array([1 / (2 ** i) for i in range(net_numpool)])

        # we don't use the lowest 2 outputs. Normalize weights so that they sum to 1
        mask = np.array([True] + [True if i < net_numpool - 1 else False for i in range(1, net_numpool)])
        weights[~mask] = 0
        weights = weights / weights.sum()
        self.ds_loss_weights = weights
        # now wrap the loss
        self.loss = MultipleOutputLoss2(self.loss, self.ds_loss_weights)
        ################# END ###################

    def forward(self, x):
        skips = []
        seg_outputs = []
        decoder_feats = []
        for d in range(len(self.conv_blocks_context) - 1):
            x = self.conv_blocks_context[d](x)
            skips.append(x)
            if not self.convolutional_pooling:
                x = self.td[d](x)

        x = self.conv_blocks_context[-1](x)

        for u in range(len(self.tu)):
            x = self.tu[u](x)
            x = torch.cat((x, skips[-(u + 1)]), dim=1)
            x = self.conv_blocks_localization[u](x)
            decoder_feats.append(x)
            seg_outputs.append(self.final_nonlin(self.seg_outputs[u](x)))

        if self._deep_supervision:
            logits = tuple([seg_outputs[-1]] + [i(j) for i, j in
                                              zip(list(self.upscale_logits_ops)[::-1], seg_outputs[:-1][::-1])])
        else:
            logits = tuple([seg_outputs[-1]])

        if self.training:
            return logits, skips, decoder_feats[::-1]
        return logits[0]




import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


class qkv_super(nn.Linear):
    def __init__(self, super_in_dim, super_out_dim, bias=True, uniform_=None, non_linear='linear', scale=False):
        super().__init__(super_in_dim, super_out_dim, bias=bias)

        # super_in_dim and super_out_dim indicate the largest network!
        self.super_in_dim = super_in_dim
        self.super_out_dim = super_out_dim

        # input_dim and output_dim indicate the current sampled size
        self.sample_in_dim = None
        self.sample_out_dim = None

        self.samples = {}

        self.scale = scale
        # self._reset_parameters(bias, uniform_, non_linear)
        self.profiling = False

    def profile(self, mode=True):
        self.profiling = mode

    def sample_parameters(self, resample=False):
        if self.profiling or resample:
            return self._sample_parameters()
        return self.samples

    def _reset_parameters(self, bias, uniform_, non_linear):
        nn.init.xavier_uniform_(self.weight) if uniform_ is None else uniform_(
            self.weight, non_linear=non_linear)
        if bias:
            nn.init.constant_(self.bias, 0.)

    def set_sample_config(self, sample_in_dim, sample_out_dim):
        self.sample_in_dim = sample_in_dim
        self.sample_out_dim = sample_out_dim

        self._sample_parameters()

    def _sample_parameters(self):
        self.samples['weight'] = sample_weight(self.weight, self.sample_in_dim, self.sample_out_dim)
        self.samples['bias'] = self.bias
        self.sample_scale = self.super_out_dim/self.sample_out_dim
        if self.bias is not None:
            self.samples['bias'] = sample_bias(self.bias, self.sample_out_dim)
        return self.samples

    def forward(self, x):
        self.sample_parameters()
        return F.linear(x, self.samples['weight'], self.samples['bias']) * (self.sample_scale if self.scale else 1)

    def calc_sampled_param_num(self):
        assert 'weight' in self.samples.keys()
        weight_numel = self.samples['weight'].numel()

        if self.samples['bias'] is not None:
            bias_numel = self.samples['bias'].numel()
        else:
            bias_numel = 0

        return weight_numel + bias_numel
    def get_complexity(self, sequence_length):
        total_flops = 0
        total_flops += sequence_length *  np.prod(self.samples['weight'].size())
        return total_flops

def sample_weight(weight, sample_in_dim, sample_out_dim):

    sample_weight = weight[:, :sample_in_dim]
    sample_weight = torch.cat([sample_weight[i:sample_out_dim:3, :] for i in range(3)], dim =0)

    return sample_weight


def sample_bias(bias, sample_out_dim):
    sample_bias = bias[:sample_out_dim]

    return sample_bias



import logging

class BasicFinetunner:
    ''' Basic finetunner

    '''
    def __init__(self,pretrained_network,is_frozen=True):
        ''' init method

        :param pretrained_network: the pretrained_network
        :param is_frozen: whether frozen the finetunned bottom layers
        '''
        self._pretrained_params = pretrained_network.state_dict()
        self.is_frozen = is_frozen

    def __str__(self):
        _str = "BasicFinetunner:\n\t"
        _str += "is_frozen:%s\n\t" % self.is_frozen
        return  _str

    def finetune_network(self,target_network):
        ''' finetune target network

        :param target_network: the target network
        :return:
        '''
        named_parameters = {name: parameter for (name, parameter) in target_network.named_parameters()}
        ############################## finetune #######################
        self.finetuned_state_keys = []
        for key in target_network.state_dict():
            if 'num_batches_tracked' in key:
                continue
            target_v = target_network.state_dict()[key]
            if key in self._pretrained_params:
                pretrained_v = self._pretrained_params[key]
                if target_v.shape == pretrained_v.shape:
                    target_v.data.copy_(pretrained_v.data)  # Copy the data of parameters
                    self.finetuned_state_keys.append(key)
                    if self.is_frozen and key in named_parameters:
                        named_parameters[key].requires_grad = False
                else:
                    logging.error('could not load layer: %s; mismatch shape: target [%s] != pretrained [%s]'%(
                        key, (target_v.shape), (pretrained_v.shape)))
                    print('could not load layer: %s; mismatch shape: target [%s] != pretrained [%s]'%(
                        key, (target_v.shape), (pretrained_v.shape)))
            else:
                logging.error('could not load layer: %s, not in pretrained'%key)
                print('could not load layer: %s, not in pretrained'%key)
        logging.info("Finetune %s keys"%len(self.finetuned_state_keys))


import torch.nn as nn
import torch
import torch.fx

from enum import Enum
class TransferStrategy(Enum):
    '''Transfer Strategy

    '''
    OnlyFinetuneStrategy = 1         # pretraining-finetuning, and the pretrained model is the same as the target model
    OnlyDistillationStrategy = 10    # distillation
    OnlyDomainAdaptionStrategy = 20  # domain adaption
    FinetuneAndDomainAdaptionStrategy = 30 # pretraining-finetuning and domain adaption
    DistillationAndDomainAdaptionStrategy = 40   # distillation and domain adaption

ALL_STRATEGIES = [
    TransferStrategy.OnlyFinetuneStrategy,
    TransferStrategy.OnlyDistillationStrategy,
    TransferStrategy.OnlyDomainAdaptionStrategy,
    TransferStrategy.FinetuneAndDomainAdaptionStrategy,
    TransferStrategy.DistillationAndDomainAdaptionStrategy
]

class TransferrableModelOutput:
    ''' TransferrableModel Output, which is composed by backbone_output,distiller_output,adapter_output

    '''
    def __init__(self,backbone_output,distiller_output,adapter_output):
        ''' Init method

        :param backbone_output: backbone output, could be none
        :param distiller_output: distiller output, could be none
        :param adapter_output: adapter output, could be none
        '''
        self.backbone_output = backbone_output
        self.distiller_output = distiller_output
        self.adapter_output = adapter_output

class TransferrableModelLoss:
    ''' TransferrableModel Loss, which is composed by total_loss,backbone_loss,distiller_loss,adapter_loss

    '''
    def __init__(self,total_loss,backbone_loss,distiller_loss,adapter_loss):
        ''' Init method

        :param total_loss: total loss
        :param backbone_loss: backbone loss, could be none
        :param distiller_loss: distiller loss, could be none
        :param adapter_loss: adapter loss, could be none
        '''
        self.total_loss = total_loss
        self.backbone_loss = backbone_loss
        self.distiller_loss = distiller_loss
        self.adapter_loss = adapter_loss

    def backward(self):
        self.total_loss.backward()

class TransferrableModel(nn.Module):
    ''' A wrapper to make model transferrable

    '''
    def __init__(self,backbone,adapter,distiller,transfer_strategy,enable_target_training_label,
                backbone_loss_weight=1.0, distiller_loss_weight=0.0, adapter_loss_weight=0.0):
        ''' Init method

        :param backbone: the backbone model to be wrapped
        :param adapter: an adapter to perform domain adaption
        :param distiller: a distiller to perform knowledge distillation
        :param transfer_strategy: transfer strategy
        :param enable_target_training_label: During training, whether use target training label or not.
        '''
        super(TransferrableModel, self).__init__()
        self.backbone = backbone

        self.adapter = adapter
        self.distiller = distiller
        self.transfer_strategy = transfer_strategy
        self.enable_target_training_label = enable_target_training_label
        self.backbone_loss_weight = backbone_loss_weight
        self.distiller_loss_weight = distiller_loss_weight
        self.adapter_loss_weight = adapter_loss_weight

        if self.transfer_strategy == TransferStrategy.OnlyFinetuneStrategy:
            if not self.enable_target_training_label:
                raise RuntimeError("Must enable target training label when only finetune.")
    
    def get_backbone(self):
        ''' Get backbone

        :return: backbone
        '''
        return self.backbone

    def __str__(self):
        _str = 'TransferrableModel:transfer_strategy = %s, enable_target_training_label = %s\n'%(
            self.transfer_strategy, self.enable_target_training_label)
        _str += "\tbackbone:%s\n"%self.backbone
        _str += "\tadapter:%s\n" % self.adapter
        _str += "\tdistiller:%s\n" % self.distiller
        return _str

    def _finetune_forward(self,x):
        ''' forward for OnlyFinetuneStrategy

        :param x: backbone input
        :return: TransferrableModelOutput
        '''
        backbone_output = self.backbone(x)
        return TransferrableModelOutput(backbone_output,None,None)

    def _finetune_loss(self,output,label):
        ''' loss for OnlyFinetuneStrategy

        :param output: backbone predict
        :param label: backbone ground truth
        :return: TransferrableModelLoss
        '''
        backbone_loss = self.backbone.loss(output, label)
        return TransferrableModelLoss(backbone_loss,backbone_loss,None,None)

    def _distillation_forward(self,x):
        ''' forward for OnlyDistillationStrategy

        :param x: input if not distiller.use_saved_logits; (input, save_values) if distiller.use_saved_logits
        :return: TransferrableModelOutput
        '''
        x_input = x[0] if self.distiller.use_saved_logits else x
        backbone_output = self.backbone(x_input)
        distiller_output = self.distiller(x)
        return TransferrableModelOutput(backbone_output, distiller_output,None)

    def _distillation_loss(self,student_output,teacher_output,student_label):
        ''' loss for OnlyDistillationStrategy

        :param student_output: student output
        :param teacher_output: teacher output
        :param student_label: student label
        :return: TransferrableModelLoss
        '''
        if self.enable_target_training_label:
            student_loss = self.backbone.loss(student_output, student_label)
        else:
            student_loss = 0.0
        
        distiller_loss = self.distiller.loss(teacher_output, student_output, target=student_label)

        total_loss = self.backbone_loss_weight * student_loss + \
                     self.distiller_loss_weight * distiller_loss

        return TransferrableModelLoss(total_loss,student_loss,distiller_loss,None)

    def _adaption_forward(self,x_tgt,x_src):
        ''' forward for OnlyDomainAdaptionStrategy or FinetuneAndDomainAdaptionStrategy

        :param x_tgt: target domain data
        :param x_src: source domain data
        :return: (TransferrableModelOutput for x_tgt, TransferrableModelOutput for x_src)
        '''
        backbone_output_tgt = self.backbone(x_tgt)
        backbone_output_src = self.backbone(x_src)
        return (
            TransferrableModelOutput(backbone_output_tgt,None, None),
            TransferrableModelOutput(backbone_output_src, None, None),
        )

    def _adaption_loss(self, input_sample, label, source_pred_to_target_pred=lambda x:x):
        ''' loss for OnlyDomainAdaptionStrategy or FinetuneAndDomainAdaptionStrategy

        :param input_sample: tuple of (source_data, target_data)
        :param label: tuple of (source_label, target_label)
        :return:  TransferrableModelLoss
        '''
        source_data, data = input_sample
        source_label, target = label

        source_output, *source_feat = self.backbone(source_data)
        source_loss = self.backbone.loss(source_output, source_label)
        
        target_loss = None
        output, *target_feat = self.backbone(data)
        output = [source_pred_to_target_pred(item) for item in output]
        if self.enable_target_training_label:
            target_loss = self.backbone.loss(output, target)
        
        adv_loss = self.adapter(*(
            (source_output, *source_feat),
            (output, *target_feat),
            source_label
        ))

        # calc total loss
        total_loss = source_loss * self.backbone_loss_weight
        if self.enable_target_training_label:
            total_loss += target_loss * self.backbone_loss_weight
        if adv_loss:
            total_loss += adv_loss

        # return (total_loss, adv_loss, source_loss, target_loss)
        return TransferrableModelLoss(total_loss,source_loss * self.backbone_loss_weight,None,adv_loss)

    def forward(self,x):
        ''' forward

        :param x: input, may be tensor or tuple
        :return: TransferrableModelOutput
        '''
        if self.transfer_strategy == TransferStrategy.OnlyFinetuneStrategy:
            x = x if isinstance(x,torch.Tensor) else x[0] # x may be a tuple
            return self._finetune_forward(x)
        elif self.transfer_strategy == TransferStrategy.OnlyDistillationStrategy:
            return self._distillation_forward(x)
        elif self.transfer_strategy == TransferStrategy.OnlyDomainAdaptionStrategy:
            if isinstance(x,torch.Tensor):
                raise RuntimeError("TransferrableModel forward for OnlyDomainAdaptionStrategy should be tuple or list, not be %s"%type(x))
            return self._adaption_forward(x[0],x[1])
        elif self.transfer_strategy == TransferStrategy.FinetuneAndDomainAdaptionStrategy:
            if isinstance(x,torch.Tensor):
                raise RuntimeError("TransferrableModel forward for FinetuneAndDomainAdaptionStrategy should be tuple or list, not be %s"%type(x))
            return self._adaption_forward(x[0],x[1])
        elif self.transfer_strategy == TransferStrategy.DistillationAndDomainAdaptionStrategy:
            if isinstance(x, torch.Tensor):
                raise RuntimeError("TransferrableModel forward for DistillationAndDomainAdaptionStrategy should be tuple or list, not be %s" % type(x))
            return self._distillation_and_adaption_forward(x[0], x[1])
        else:
            raise RuntimeError("Unknown transfer_strategy [%s] " % self.transfer_strategy)

    def loss(self,output,label):
        ''' loss function

        :param output: output of forward()
        :param label: label (may be a tuple)
        :return: TransferrableModelLoss
        '''
        if self.transfer_strategy == TransferStrategy.OnlyFinetuneStrategy:
            return self._finetune_loss(output.backbone_output,label if isinstance(label, torch.Tensor) else label[0])
        elif self.transfer_strategy == TransferStrategy.OnlyDistillationStrategy:
            label = label if isinstance(label, torch.Tensor) else label[0]
            return self._distillation_loss(output.backbone_output,output.distiller_output,label)
        elif self.transfer_strategy == TransferStrategy.OnlyDomainAdaptionStrategy or \
            self.transfer_strategy == TransferStrategy.FinetuneAndDomainAdaptionStrategy:
            return self._adaption_loss(output, label)
        elif self.transfer_strategy == TransferStrategy.DistillationAndDomainAdaptionStrategy:
            return self._distillation_and_adaption_loss(
                                        output[0].backbone_output, output[1].backbone_output,
                                        output[1].distiller_output, output[1].distiller_output,
                                        output[0].adapter_output,output[1].adapter_output,
                                        label[0], label[1])
        else:
            raise RuntimeError("Unknown transfer_strategy [%s] " % self.transfer_strategy)

    def get_training_metrics(self,output,label,loss_value,metric_fn_map):
        ''' get the training metrics

        :param output: output of forward()
        :param label: label (may be a tuple)
        :param loss_value: output of loss()
        :param metric_fn_map:  metric function map, which map metric name to metric function
        :return: metrics
        '''
        metric_values = {"total_loss": loss_value.total_loss}
        if loss_value.backbone_loss is not None:
            metric_values["backbone_loss"] = loss_value.backbone_loss
        if loss_value.distiller_loss is not None:
            metric_values["distiller_loss"] = loss_value.distiller_loss
        if loss_value.adapter_loss is not None:
            metric_values["adapter_loss"] = loss_value.adapter_loss

        for (metric_name, metric_fn) in sorted(metric_fn_map.items()):
            if self.transfer_strategy in [
                TransferStrategy.OnlyDomainAdaptionStrategy,
                TransferStrategy.FinetuneAndDomainAdaptionStrategy,
                TransferStrategy.DistillationAndDomainAdaptionStrategy]:
                if (isinstance(label, torch.Tensor)):
                    raise RuntimeError("label in adaption must be a collection, not %s" % (type(label)))

                # 0 is target domain output, 1 is source domain output
                metric_value = metric_fn(output[1].backbone_output, label[1])
                metric_value = metric_value[0] if isinstance(metric_value, list) else metric_value
                metric_values["%s_src_domain"%metric_name] = metric_value
                if self.enable_target_training_label:
                    metric_value = metric_fn(output[0].backbone_output, label[0])
                    metric_value = metric_value[0] if isinstance(metric_value, list) else metric_value
                    metric_values["%s_target_domain" % metric_name] = metric_value
            else:
                metric_value = metric_fn(output.backbone_output, label if isinstance(label, torch.Tensor) else label[0])
                metric_value = metric_value[0] if isinstance(metric_value, list) else metric_value
                metric_values[metric_name] = metric_value

        return metric_values


def _make_transferrable(model, loss,
                        finetunner, distiller, adapter,
                        transfer_strategy, enable_target_training_label,
                        backbone_loss_weight=1.0, distiller_loss_weight=0.0, adapter_loss_weight=0.0):
    ''' make a model transferrable

    :param model: the backbone model. If model does not have loss method, then use loss argument.
    :param loss : loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param finetunner: a finetunner. If no finetune, then could be none.
    :param distiller: a distiller. If no distillation, then could be none.
    :param adapter: an adapter. If no adaption, then could be none.
    :param transfer_strategy: transfer strategy.
    :param enable_target_training_label: During training, whether use target training label or not.
    :param backbone_loss_weight: the weight of backbone model loss, default=1
    :param distiller_loss_weight: the weight of distiller_loss_weight, default=0
    :param adapter_loss_weight: the weight of adapter_loss_weight, default=0
    :return: a TransferrableModel
    '''
    ######## check input #####
    if not hasattr(model,"loss"):
        if loss is None:
            raise RuntimeError("Need loss for model")
        else:
            setattr(model,'loss',loss)
    if transfer_strategy in [TransferStrategy.OnlyDistillationStrategy,
                             TransferStrategy.DistillationAndDomainAdaptionStrategy]:
        if distiller is None:
            raise RuntimeError("Need distiller for Distillation")
    if transfer_strategy in [TransferStrategy.OnlyDomainAdaptionStrategy,
                             TransferStrategy.FinetuneAndDomainAdaptionStrategy,
                             TransferStrategy.DistillationAndDomainAdaptionStrategy]:
        if adapter is None:
            raise RuntimeError("Need adapter for Adaption")
    if transfer_strategy == TransferStrategy.OnlyFinetuneStrategy:
        if not enable_target_training_label:
            raise RuntimeError("Need enable_target_training_label for OnlyFinetune")
    ###################### initialize ####################
    if transfer_strategy in [TransferStrategy.OnlyFinetuneStrategy,
                                      TransferStrategy.FinetuneAndDomainAdaptionStrategy]:
        finetunner.finetune_network(model)
    
    return TransferrableModel(model,adapter,distiller,transfer_strategy,enable_target_training_label,backbone_loss_weight,distiller_loss_weight,adapter_loss_weight)

def _transferrable(loss, finetunner, distiller, adapter,
                   transfer_strategy, enable_target_training_label,
                   backbone_loss_weight=1.0, distiller_loss_weight=0.0, adapter_loss_weight=0.0):
    ''' a decorator to make instances of class transferrable

    :param loss: loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param finetunner: a finetunner. If no finetune, then could be none.
    :param distiller: a distiller. If no distillation, then could be none.
    :param adapter: an adapter. If no adaption, then could be none.
    :param transfer_strategy: transfer strategy
    :param enable_target_training_label: During training, whether use target training label or not.
    :return: a wrapper
    '''
    def wrapper(ModelClass):
        def _wrapper(*args, **kargs):
            model = ModelClass(*args, **kargs)
            return _make_transferrable(model, loss,
                                       finetunner, distiller, adapter,
                                       transfer_strategy, enable_target_training_label,
                                       backbone_loss_weight,distiller_loss_weight,adapter_loss_weight)
        return _wrapper
    return wrapper

############ simple API #############
def make_transferrable_with_finetune(model,loss,finetunner):
    ''' make transferrable with finetune strategy

    :param model: the backbone model. If model does not have loss method, then use loss argument.
    :param loss : loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param finetunner: a finetunner.
    :return: a TransferrableModel
    '''
    return _make_transferrable(model=model, loss=loss,
                               finetunner=finetunner, distiller=None, adapter=None,
                               transfer_strategy=TransferStrategy.OnlyFinetuneStrategy,
                               enable_target_training_label=True)

def transferrable_with_finetune(loss,finetunner):
    ''' a decorator to make instances of class transferrable with finetune strategy

    :param loss : loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param finetunner: a finetunner.
    :return: a wrapper
    '''
    return _transferrable(loss=loss,
                          finetunner=finetunner, distiller=None, adapter=None,
                          transfer_strategy=TransferStrategy.OnlyFinetuneStrategy,
                          enable_target_training_label=True)

def make_transferrable_with_knowledge_distillation(model,loss,distiller,
                                                   enable_target_training_label=True,backbone_loss_weight=0.1,distiller_loss_weight=0.9):
    '''  make transferrable with knowledge distillation strategy

    :param model: the backbone model. If model does not have loss method, then use loss argument.
    :param loss : loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param distiller: a distiller.
    :param enable_target_training_label: During training, whether use target training label or not.
    :return: a TransferrableModel
    '''
    return _make_transferrable(model=model, loss=loss,
                               finetunner=None, distiller=distiller, adapter=None,
                               transfer_strategy=TransferStrategy.OnlyDistillationStrategy,
                               enable_target_training_label=enable_target_training_label,
                               backbone_loss_weight=backbone_loss_weight,distiller_loss_weight=distiller_loss_weight)

def transferrable_with_knowledge_distillation(loss,distiller,
                                                   enable_target_training_label,backbone_loss_weight,distiller_loss_weight):
    ''' a decorator to make instances of class transferrable with distillation strategy

    :param loss : loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param distiller: a distiller.
    :param enable_target_training_label: During training, whether use target training label or not.
    :return: a wrapper
    '''
    return _transferrable(loss=loss,
                          finetunner=None, distiller=distiller, adapter=None,
                          transfer_strategy=TransferStrategy.OnlyDistillationStrategy,
                          enable_target_training_label=enable_target_training_label,
                          backbone_loss_weight=backbone_loss_weight,distiller_loss_weight=distiller_loss_weight)

def make_transferrable_with_domain_adaption(model,loss,adapter,
                                            enable_target_training_label,backbone_loss_weight,adapter_loss_weight):
    ''' make transferrable with domain adaption strategy

    :param model: the backbone model. If model does not have loss method, then use loss argument.
    :param loss : loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param adapter: an adapter.
    :param enable_target_training_label:  During training, whether use target training label or not.
    :return: a TransferrableModel
    '''
    return _make_transferrable(model=model, loss=loss,
                               finetunner=None, distiller=None, adapter=adapter,
                               transfer_strategy=TransferStrategy.OnlyDomainAdaptionStrategy,
                               enable_target_training_label=enable_target_training_label,
                               backbone_loss_weight=backbone_loss_weight,adapter_loss_weight=adapter_loss_weight)

def transferrable_with_domain_adaption(loss,adapter,
                                        enable_target_training_label,backbone_loss_weight,adapter_loss_weight):
    ''' a decorator to make instances of class transferrable with adaption strategy

    :param loss : loss function for model,signature: loss(output_logit, label). If model has loss attribute, then loss could be none.
    :param adapter: an adapter.
    :param enable_target_training_label:  During training, whether use target training label or not.
    :return: a wrapper
    '''
    return _transferrable(loss=loss,
                          finetunner=None, distiller=None, adapter=adapter,
                          transfer_strategy=TransferStrategy.OnlyDomainAdaptionStrategy,
                          enable_target_training_label=enable_target_training_label,
                          backbone_loss_weight=backbone_loss_weight,adapter_loss_weight=adapter_loss_weight)



import os
import torch
import e2eAIOK.common.trainer.utils.extend_distributed as ext_dist
# from e2eAIOK.common.trainer.data.data_utils.data_utils import channels_last_collate
from e2eAIOK.common.trainer.data_builder import DataBuilder

class DataBuilderCV(DataBuilder):
    def __init__(self, cfg):
        super().__init__(cfg)
    
    def get_dataloader(self):
        """
            create training/evaluation/test dataloader
        """
        ### generate datasets
        if self.dataset_train is None or self.dataset_val is None:
            self.prepare_dataset()

        ### get configurations
        drop_last =  self.cfg.drop_last if "drop_last" in self.cfg else False
        pin_memory = self.cfg.pin_mem if "pin_mem" in self.cfg else False

        ### check whether distributed or enable_ipex
        if ext_dist.my_size > 1:
            num_tasks = ext_dist.dist.get_world_size()
            global_rank = ext_dist.dist.get_rank()
            sampler_train = torch.utils.data.DistributedSampler(
                self.dataset_train, num_replicas=num_tasks, rank=global_rank, shuffle=True, drop_last= drop_last)
            sampler_val = torch.utils.data.DistributedSampler(
                self.dataset_val, num_replicas=num_tasks, rank=global_rank,  shuffle=False)
            if self.dataset_test is not None:
                sampler_test = torch.utils.data.DistributedSampler(
                self.dataset_test, num_replicas=num_tasks, rank=global_rank,  shuffle=False)
        else:
            sampler_train = torch.utils.data.RandomSampler(self.dataset_train)
            sampler_val = torch.utils.data.SequentialSampler(self.dataset_val)
            sampler_test = torch.utils.data.SequentialSampler(self.dataset_test)
        
        # if "enable_ipex" in self.cfg and self.cfg.enable_ipex:
        #     collate_fn = channels_last_collate
        # else:
        collate_fn = None

        ### generate dataloader
        dataloader_train = torch.utils.data.DataLoader(
            self.dataset_train, 
            sampler=sampler_train,
            batch_size=self.cfg.train_batch_size,
            num_workers=self.cfg.num_workers,
            drop_last=drop_last,
            collate_fn=collate_fn,
            pin_memory=pin_memory
        )

        dataloader_val = torch.utils.data.DataLoader(
            self.dataset_val, 
            batch_size=self.cfg.eval_batch_size,
            sampler=sampler_val, 
            num_workers=self.cfg.num_workers,
            drop_last=False,
            collate_fn=collate_fn
        )

        if self.dataset_test is None:
            return dataloader_train, dataloader_val
        else:
            dataloader_test = torch.utils.data.DataLoader(
                self.dataset_test, 
                batch_size=self.cfg.eval_batch_size,
                sampler=sampler_test, 
                num_workers=self.cfg.num_workers,
                drop_last=False,
                collate_fn=collate_fn
            )
        return dataloader_train, dataloader_val, dataloader_test